/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * transport.c
 * -----------
 *
 * Inter-process communication architecture
 */

#include "../../../config.h"

#include "impulse/comms/main.h"
#include "impulse/util/memory.h"
#include "impulse/util/error.h"
#include "impulse/util/hash.h"
#include "impulse/util/thread.h"


/* 
 * The standard builtin communication transports
 */
#include "impulse/comms/transport_sock.h"
#include "impulse/comms/transport_ipc.h"


static impulse_hash transports = NULL;
#if USE_THREADS
static impulse_rwlock_static transports_lock = IMPULSE_RWLOCK_CREATE_STATIC;
#endif


void impulse_comms_init(void)
{
#if USE_THREADS
  impulse_rwlock_write(&transports_lock);
#endif
  if (transports)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot initialise communications subsystem",8);
    impulse_unistring msg = impulse_unistring_create_utf("Communications subsystem has already been initialised",8);
#if USE_THREADS
    impulse_rwlock_unlock(&transports_lock);
#endif
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, msg);
    impulse_unistring_destroy(desc);
    impulse_unistring_destroy(msg);
  }
  transports = impulse_hash_create();

#if HAVE_COMMS_SOCK
  impulse_hash_store(transports, impulse_transport_sock->name, impulse_transport_sock);
#endif
#if HAVE_COMMS_IPC
  impulse_hash_store(transports, impulse_transport_ipc->name, impulse_transport_ipc);
#endif
  
#if USE_THREADS
  impulse_rwlock_unlock(&transports_lock);
#endif
}


void impulse_comms_cleanup(void)
{
#if USE_THREADS
  impulse_rwlock_write(&transports_lock);
#endif
  if (!transports)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot cleanup communications subsystem",8);
    impulse_unistring msg = impulse_unistring_create_utf("Communications subsystem has not been initialised",8);
#if USE_THREADS
    impulse_rwlock_unlock(&transports_lock);
#endif
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, msg);
    impulse_unistring_destroy(desc);
    impulse_unistring_destroy(msg);
  }
  impulse_hash_destroy(transports);

  transports = NULL;
#if USE_THREADS
  impulse_rwlock_unlock(&transports_lock);
#endif
}


void impulse_comms_register(impulse_transport transport)
{
#if USE_THREADS
  impulse_rwlock_write(&transports_lock);
#endif
  if (!transports)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot register transport protocol",8);
    impulse_unistring msg = impulse_unistring_create_utf("Communications subsystem has not been initialised",8);
#if USE_THREADS
    impulse_rwlock_unlock(&transports_lock);
#endif
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, msg);
    impulse_unistring_destroy(desc);
    impulse_unistring_destroy(msg);
  }
  impulse_hash_store(transports, transport->name, transport);
#if USE_THREADS
  impulse_rwlock_unlock(&transports_lock);
#endif
}


void impulse_comms_deregister(impulse_transport transport)
{
#if USE_THREADS
  impulse_rwlock_write(&transports_lock);
#endif
  if (!transports)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot deregister transport protocol",8);
    impulse_unistring msg = impulse_unistring_create_utf("Communications subsystem has not been initialised",8);
#if USE_THREADS
    impulse_rwlock_unlock(&transports_lock);
#endif
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, msg);
    impulse_unistring_destroy(desc);
    impulse_unistring_destroy(msg);
  }
  impulse_hash_remove(transports, transport->name);
#if USE_THREADS
  impulse_rwlock_unlock(&transports_lock);
#endif
}


impulse_transport impulse_comms_lookup(impulse_unistring name)
{
  impulse_transport transport = NULL;
#if USE_THREADS
  impulse_rwlock_read(&transports_lock);
#endif
  if (!transports)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Canno lookupr transport protocol",8);
    impulse_unistring msg = impulse_unistring_create_utf("Communications subsystem has not been initialised",8);
#if USE_THREADS
    impulse_rwlock_unlock(&transports_lock);
#endif
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, msg);
    impulse_unistring_destroy(desc);
    impulse_unistring_destroy(msg);
  }
  transport = impulse_hash_recover(transports, name);

#if USE_THREADS
  impulse_rwlock_unlock(&transports_lock);
#endif
  return transport;
}


impulse_int impulse_comms_client(impulse_unistring address_local,
				 impulse_unistring address_remote,
				 impulse_client *client)
{
  impulse_int offset = impulse_unistring_find_chr(address_remote, (impulse_unichar)':');
  impulse_unistring proto = NULL;
  impulse_transport transport = NULL;
  void *data = NULL;
  
  if (offset < 0)
    return -1;

  proto = impulse_unistring_create_empty();
  impulse_unistring_substring(proto, address_remote, 0, offset);
  transport = impulse_comms_lookup(proto);
  impulse_unistring_destroy(proto);
  
  if (!transport)
    return -1;
  
  data = transport->client.create();
  
  *client = impulse_client_create(transport, data);
  
  return impulse_client_open(*client, address_local, address_remote);
}


impulse_int impulse_comms_server(impulse_unistring address, 
				 impulse_server *server)
{
  impulse_int offset = impulse_unistring_find_chr(address, (impulse_unichar)':');
  impulse_unistring proto = NULL;
  impulse_transport transport = NULL;
  void *data = NULL;
  
  if (offset < 0)
    return -1;
  
  proto = impulse_unistring_create_empty();
  impulse_unistring_substring(proto, address, 0, offset);
  transport = impulse_comms_lookup(proto);
  impulse_unistring_destroy(proto);

  if (!transport)
    return -1;
 
  data = transport->server.create();
  
  *server = impulse_server_create(transport, data);
  
  return impulse_server_open(*server, address);
}


impulse_int impulse_comms_peer(impulse_unistring address, 
			       impulse_peer *peer)
{
  impulse_int offset = impulse_unistring_find_chr(address, (impulse_unichar)':');
  impulse_unistring proto = NULL;
  impulse_transport transport = NULL;
  void *data = NULL;
  
  if (offset < 0)
    return -1;
  
  proto = impulse_unistring_create_empty();
  impulse_unistring_substring(proto, address, 0, offset);
  transport = impulse_comms_lookup(proto);
  impulse_unistring_destroy(proto);

  if (!transport)
    return -1;
  
  data = transport->peer.create();
  
  *peer = impulse_peer_create(transport, data);
  
  return impulse_peer_open(*peer, address);
}


/* End of file */
