/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * transport_sock.c
 * ----------------
 *
 * Communication transport functions
 * Sockets protocols
 * 
 * Addresses are specified in the form:
 * 
 * protocol:node:service
 * 
 * For IP based services, 'node' can be a host name,
 * IPv4 dotted decimal address, or IPv6 colon separated
 * hexidecimal address. 'service' is either a numeric
 * port number or service name from /etc/services or the like
 * 
 * For Unix domain sockets, 'node' must be one of 'unix' or
 * 'local', and 'service' is a full path name.
 * 
 * eg.
 * 
 * sock:www.google.com:http
 * sock:10.0.0.1:ftp
 * sock:localhost:6000
 * sock:unix:/var/run/mysocket
 * sock:f340:f353::f454:telnet
 * 
 */

#include "../../../config.h"

#include "impulse/comms/transport_sock.h"
#include "impulse/util/memory.h"
#include "impulse/util/error.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#if HAVE_COMMS_SOCK

#include <netdb.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/ioctl.h>

/*                            ======--------======                           */


struct _impulse_transport_socks
{
  impulse_int handle;
  
  impulse_char *node;
  impulse_char *service;
};

struct _impulse_transport_sockc
{
  impulse_int handle;

  impulse_char *node_local;
  impulse_char *node_remote;
  impulse_char *service_local;
  impulse_char *service_remote;
};

struct _impulse_transport_sockp
{
  impulse_int handle;
  
  impulse_char *node;
  impulse_char *service;
};


typedef struct _impulse_transport_socks *impulse_transport_socks;
typedef struct _impulse_transport_sockc *impulse_transport_sockc;
typedef struct _impulse_transport_sockp *impulse_transport_sockp;


union sockaddr_all
{
  struct sockaddr a;
#if HAVE_COMMS_SOCK_UNIX || HAVE_COMMS_SOCK_LOCAL
  struct sockaddr_un un;
#endif
#if HAVE_COMMS_SOCK_IPV4
  struct sockaddr_in in;
#endif
#if HAVE_COMMS_SOCK_IPV6
  struct sockaddr_in6 in6;
#endif
};


/*                            ======--------======                           */

/* Server functions */
void *impulse_transport_socks_create(void);
void impulse_transport_socks_destroy(void *data);

impulse_bool impulse_transport_socks_is_open(void *data);
impulse_int impulse_transport_socks_open(void *data,
					 impulse_unistring address);
impulse_int impulse_transport_socks_close(void *data);
impulse_int impulse_transport_socks_handle(void *data);

impulse_int impulse_transport_socks_pending(void *data);
impulse_int impulse_transport_socks_accept(void *data,
					   void **client);
impulse_int impulse_transport_socks_address(void *data,
					    impulse_unistring *address);


/* Client functions */
void *impulse_transport_sockc_create(void);
void *impulse_transport_sockc_create_handle(impulse_int handle);
void impulse_transport_sockc_destroy(void *data);

impulse_bool impulse_transport_sockc_is_open(void *data);
impulse_int impulse_transport_sockc_open(void *data,
					 impulse_unistring address_local,
					 impulse_unistring address_remote);
impulse_int impulse_transport_sockc_close(void *data);
impulse_int impulse_transport_sockc_handle(void *data);

impulse_int impulse_transport_sockc_pending(void *data);
impulse_int impulse_transport_sockc_read(void *data,
					 impulse_byte *buffer,
					 impulse_int length);
impulse_int impulse_transport_sockc_write(void *data,
					  impulse_byte *buffer,
					  impulse_int length);

impulse_int impulse_transport_sockc_address_local(void *data,
						  impulse_unistring *address);
impulse_int impulse_transport_sockc_address_remote(void *data,
						   impulse_unistring *address);

/* Peer functions */
void *impulse_transport_sockp_create(void);
void impulse_transport_sockp_destroy(void *data);

impulse_bool impulse_transport_sockp_is_open(void *data);
impulse_int impulse_transport_sockp_open(void *data,
					 impulse_unistring address);
impulse_int impulse_transport_sockp_close(void *data);
impulse_int impulse_transport_sockp_handle(void *data);

impulse_int impulse_transport_sockp_pending(void *data);
impulse_int impulse_transport_sockp_send(void *data,
					 impulse_message msg,
					 impulse_unistring address);
impulse_int impulse_transport_sockp_receive(void *data,
					    impulse_message *msg,
					    impulse_unistring *address);

impulse_int impulse_transport_sockp_address(void *data,
					    impulse_unistring *address);

/*                            ======--------======                           */


/* Static unicode strings for protocol name */
static struct _impulse_unistring _name_sock =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'s',
    (impulse_unichar)'o',
    (impulse_unichar)'c', 
    (impulse_unichar)'k', 
    (impulse_unichar)'\0'
  }
};

/* Transport function blocks */

static struct _impulse_transport _impulse_transport_sock =
{
  &_name_sock,
  {
    impulse_transport_socks_create,
    impulse_transport_socks_destroy,
    impulse_transport_socks_is_open,
    impulse_transport_socks_open,
    impulse_transport_socks_close,
    impulse_transport_socks_handle,
    impulse_transport_socks_pending,
    impulse_transport_socks_accept,
    impulse_transport_socks_address,
  },
  {
    impulse_transport_sockc_create,
    impulse_transport_sockc_destroy,
    impulse_transport_sockc_is_open,
    impulse_transport_sockc_open,
    impulse_transport_sockc_close,
    impulse_transport_sockc_handle,
    impulse_transport_sockc_pending,
    impulse_transport_sockc_read,
    impulse_transport_sockc_write,
    impulse_transport_sockc_address_remote,
    impulse_transport_sockc_address_remote
  },
  {
    impulse_transport_sockp_create,
    impulse_transport_sockp_destroy,
    impulse_transport_sockp_is_open,
    impulse_transport_sockp_open,
    impulse_transport_sockp_close,
    impulse_transport_sockp_handle,
    impulse_transport_sockp_pending,
    impulse_transport_sockp_send,
    impulse_transport_sockp_receive,
    impulse_transport_sockp_address
  }
};

impulse_transport impulse_transport_sock = &_impulse_transport_sock;


/*                            ======--------======                           */

/* Miscellaneous helper functions */

static impulse_int impulse_transport_sock_split(impulse_unistring address,
						impulse_char **node,
						impulse_char **service)
{
  impulse_int offset1 = impulse_unistring_find_chr(address, (impulse_unichar)':');
  impulse_int offset2 = impulse_unistring_rfind_chr(address, (impulse_unichar)':');
  impulse_unistring unode = NULL;
  impulse_unistring uservice = NULL; 

  if (*node)
    impulse_free(node);
  if (*service)
    impulse_free(service);
  
  if (offset1 < 0 || offset2 < 0 ||
      offset1 == offset2)
  {
    *node = NULL;
    *service = NULL;
    return -1;
  }

  /* First the node */
  unode = impulse_unistring_create_empty();
  impulse_unistring_substring(unode, address, offset1+1, offset2);
  *node = impulse_unistring_to_utf(unode, 8, 0);
  impulse_unistring_destroy(unode);

  /* And now the service */
  uservice = impulse_unistring_create_empty();
  impulse_unistring_substring(uservice, address, offset2+1, address->length);
  *service = impulse_unistring_to_utf(uservice, 8, 0);
  impulse_unistring_destroy(uservice);

  return 0;
}


static impulse_int impulse_transport_sock_join(impulse_char *node,
					       impulse_char *service,
					       impulse_unistring *address)
{
  impulse_int length = strlen(node) + strlen(service) + 10;
  impulse_char *buffer = impulse_alloc(length * sizeof(impulse_char));
  
  snprintf(buffer, length-1, "sock:%s:%s", node, service);
  *address = impulse_unistring_create_utf(buffer, 8);
  impulse_free(buffer);
  
  return 0;
}


static impulse_int impulse_transport_sock_query(union sockaddr_all *addr,
						impulse_int addr_len,
						impulse_char **node,
						impulse_char **service,
						impulse_int type)
{
  if (*node)
    impulse_free(*node);
  if (*service)
    impulse_free(*service);
  *node = NULL;
  *service = NULL;
  
  switch (addr->a.sa_family)
  {
#if HAVE_COMMS_SOCK_LOCAL
   case AF_LOCAL:
#elif HAVE_COMMS_SOCK_UNIX
   case AF_UNIX:
#endif
#if HAVE_COMMS_SOCK_LOCAL || HAVE_COMMS_SOCK_UNIX
    {
      char *node_buf = "local";
      *node = impulse_alloc(sizeof(impulse_char)*(strlen(node_buf)+1));
      strcpy(*node, node_buf);
      *service = impulse_alloc(sizeof(impulse_char)*(strlen(addr->un.sun_path)+1));
      strcpy(*service, addr->un.sun_path);
    }
    break;
#endif
    
#if HAVE_COMMS_SOCK_ADDRINFO
# if HAVE_COMMS_SOCK_IPV4
   case AF_INET:
# endif
# if HAVE_COMMS_SOCK_IPV6
   case AF_INET6:
# endif
# if HAVE_COMMS_SOCK_IPV4 || HAVE_COMMS_SOCK_IPV6
    {
      impulse_int ret = 0;
      impulse_char node_buf[NI_MAXHOST];
      impulse_char service_buf[NI_MAXSERV];
  
      if ((ret = getnameinfo(&addr->a, addr_len, 
			     node_buf, NI_MAXHOST, 
			     service_buf, NI_MAXSERV, (type == SOCK_DGRAM) ? NI_DGRAM : 0)) < 0)
      {
	/* do something */
      }
    }
    break;
# endif
#else

# if HAVE_COMMS_SOCK_IPV4
   case AF_INET:
    {
      struct hostent *he = NULL;
      impulse_short port;
      impulse_char port_buf[50];
      
      if (!(he = gethostbyaddr((char *) &addr->in.sin_addr, sizeof(addr->in.sin_addr), AF_INET)))
	return -1;
      port = ntohs(addr->in.sin_port);
      sprintf(port_buf, "%d", port);
      *service = impulse_alloc(sizeof(impulse_char)*(strlen(port_buf)+1));
      strcpy(*service, port_buf);
      
      *node = impulse_alloc(sizeof(impulse_char)*(strlen(he->h_name)+1));
      strcpy(*node, he->h_name);
    }
    break;
# endif
    
# if HAVE_COMMS_SOCK_IPV6
   case AF_INET6:
    {
      struct hostent *he = NULL;
      impulse_short port;
      impulse_char port_buf[50];
      
      if (!(he = gethostbyaddr((char *) &addr->in6.sin6_addr, sizeof(addr->in6.sin6_addr), AF_INET6)))
	return -1;
      port = ntohs(addr->in6.sin6_port);
      sprintf(port_buf, "%d", port);
      *service = impulse_alloc(sizeof(impulse_char)*(strlen(port_buf)+1));
      strcpy(*service, port_buf);
      
      *node = impulse_alloc(sizeof(impulse_char)*(strlen(he->h_name)+1));
      strcpy(*node, he->h_name);
    }
    break;
# endif
#endif    
   default:
    {
      impulse_unistring desc = impulse_unistring_create_utf("Lookup of unknown address family requested", 8);
      impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, NULL);
      impulse_unistring_destroy(desc);
    }
  }
  
  return 0;
}


static impulse_int impulse_transport_sock_lookup(impulse_unistring address,
						 impulse_char **node,
						 impulse_char **service,
						 union sockaddr_all **addr,
						 impulse_int *addr_len,
						 impulse_int type)
{
#if HAVE_COMMS_SOCK_ADDRINFO
  impulse_int ret = 0;
  struct addrinfo in;
  struct addrinfo *out;
  
  *addr = NULL;
  *addr_len = 0;
  
  if (impulse_transport_sock_split(address,
				   node,
				   service) < 0)
    return -1;

  in.ai_flags = 0;
  in.ai_socktype = type;
  in.ai_family = 0;
  in.ai_protocol = 0;
  in.ai_addrlen = 0;
  in.ai_canonname = NULL;
  in.ai_addr = NULL;
  in.ai_next = NULL;
  
  if ((ret = getaddrinfo(*node, *service, &in, &out)) != 0)
  {
    impulse_free(*node);
    impulse_free(*service);
    *node = NULL;
    *service = NULL;
    return -1;
  }
  
  /*
   * Fix me: we should really walk through the linked
   * list and pick out the best entry.
   */
  *addr_len = out->ai_addrlen;
  *addr = impulse_alloc(*addr_len);
  memcpy(*addr, out->ai_addr, *addr_len);
  
  freeaddrinfo(out);
#else
  *addr = NULL;
  *addr_len = 0;
  
  if (impulse_transport_sock_split(address,
				   node,
				   service) < 0)
    return -1;
  
  if (!strcmp(*node, "local") ||
      !strcmp(*node, "unix"))
  {
    *addr = impulse_alloc(sizeof(union sockaddr_all));
# if HAVE_COMMS_SOCK_LOCAL
    (*addr)->un.sun_family = AF_LOCAL;
# elif HAVE_COMMS_SOCK_UNIX
    (*addr)->un.sun_family = AF_UNIX;
# else
    impulse_free(*node);
    impulse_free(*service);
    *node = NULL;
    *service = NULL;
    impulse_free(*addr);
    *addr = NULL;
    return -1;
# endif
    strncpy((*addr)->un.sun_path, *service, 107);
    (*addr)->un.sun_path[107] = '\0';
    *addr_len = SUN_LEN(&(*addr)->un);
  }
  else
  {
    struct hostent *he = NULL;
    struct servent *se = NULL;
    impulse_int port = 0;

    /*
     * Fix me: replace with reentrant ones if available
     * or make this stuff thread safe with mutexes
     */
    if (strlen(*node) && !(he = gethostbyname(*node)))
    {
      impulse_free(*node);
      impulse_free(*service);
      *node = NULL;
      *service = NULL;
      return -1;
    }
    if (strlen(*service))
    {
      impulse_char *end = NULL;
      port = strtoul(*service, &end, 10);
      if (end == *service)
      {
	if (!(se = getservbyname(*service, (type == SOCK_STREAM)?"tcp" : "udp")))
	{
	  impulse_free(*node);
	  impulse_free(*service);
	  *node = NULL;
	  *service = NULL;
	  return -1;
	}
	port = se->s_port;
      }
      else
	port = htons(port);
    }
    
    switch (he->h_addrtype)
    {
# if HAVE_COMMS_SOCK_IPV4
     case AF_INET:
      {
	(*addr) = impulse_alloc(sizeof(union sockaddr_all));
	*addr_len = sizeof(union sockaddr_all);
	
	(*addr)->in.sin_family = AF_INET;
	  (*addr)->in.sin_port = port;
	if (!strlen(*node))
	  (*addr)->in.sin_addr.s_addr = INADDR_ANY;
	else
	  (*addr)->in.sin_addr = *((struct in_addr *)he->h_addr);
	memset(&((*addr)->in.sin_zero), 0, 8);
      }
      break;
# endif
    
# if HAVE_COMMS_SOCK_IPV6
     case AF_INET6:
      {
	(*addr) = impulse_alloc(sizeof(union sockaddr_all));
	*addr_len = sizeof(union sockaddr_all);

	(*addr)->in6.sin6_family = AF_INET6;
	(*addr)->in6.sin6_port = port;
	(*addr)->in6.sin6_flowinfo = 0;
	if (!strlen(*node))
	  (*addr)->in6.sin6_addr = in6addr_any;
	else
	  (*addr)->in6.sin6_addr = *((struct in6_addr *)he->h_addr);
	/* 
	 * This field is mentioned in RFC 2553, but does not seem to
	 * be implemented?
	 */
	//addr->in6.sin6_scope_id = 0;
      }
      break;
# endif
      
     default:
      impulse_free(*node);
      impulse_free(*service);
      *node = NULL;
      *service = NULL;
      return -1;
    }
  }
#endif
  return 0;
}


/* Server functions */

void *impulse_transport_socks_create(void)
{
  impulse_transport_socks server = impulse_alloc(sizeof(struct _impulse_transport_socks));
  
  server->handle = -1;

  server->service = NULL;
  server->node = NULL;
  
  return server;
}


void impulse_transport_socks_destroy(void *data)
{
  impulse_transport_socks server = (impulse_transport_socks)data;
  
  if (server->handle != -1)
    impulse_transport_socks_close(data);

  if (server->node)
    impulse_free(server->node);
  if (server->service)
    impulse_free(server->service);
  
  impulse_free(server);
}


impulse_bool impulse_transport_socks_is_open(void *data)
{
  impulse_transport_socks server = (impulse_transport_socks)data;
  
  return server->handle != -1;
}


impulse_int impulse_transport_socks_open(void *data,
					 impulse_unistring address)
{
  impulse_transport_socks server = (impulse_transport_socks)data;
  impulse_uint reuseaddr = 1;
  union sockaddr_all *addr;
  impulse_int addr_len;

  if (server->handle != -1)
    impulse_transport_socks_close(data);

  if (impulse_transport_sock_lookup(address,
				    &server->node,
				    &server->service,
				    &addr,
				    &addr_len,
				    SOCK_STREAM) < 0)
    return -1;

  if ((server->handle = socket(addr->a.sa_family,
			       SOCK_STREAM, 0)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot create server socket", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    /* Criticals never return */
  }

  if (setsockopt(server->handle, SOL_SOCKET, SO_REUSEADDR, (void *)&reuseaddr, sizeof(reuseaddr)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot reuse server socket address", 8);
    impulse_error_syscall(IMPULSE_ERROR_WARNING, desc);
    impulse_unistring_destroy(desc);
  }

  if (bind(server->handle, &addr->a, addr_len) < 0)
  {
    impulse_free(addr);
    close(server->handle);
    server->handle = -1;
    impulse_free(server->node);
    impulse_free(server->service);
    server->node = NULL;
    server->service = NULL;
    if (errno != EADDRINUSE)
    {
      impulse_unistring desc = impulse_unistring_create_utf("Cannot bind to server address", 8);
      impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
      impulse_unistring_destroy(desc);
    }
    return -1;
  }
  
  if (listen(server->handle, 20) < 0)
  {
    impulse_free(addr);
    close(server->handle);
    server->handle = -1;
    impulse_free(server->node);
    impulse_free(server->service);
    server->node = NULL;
    server->service = NULL;
    return -1;
  }

  return 0;
}


impulse_int impulse_transport_socks_close(void *data)
{
  impulse_transport_socks server = (impulse_transport_socks)data;
  
  if (server->handle != -1)
    close(server->handle);
  server->handle = -1;
  impulse_free(server->node);
  impulse_free(server->service);
  server->node = NULL;
  server->service = NULL;

  return 0;
}


impulse_int impulse_transport_socks_handle(void *data)
{
  impulse_transport_socks server = (impulse_transport_socks)data;
  
  return server->handle;
}


impulse_int impulse_transport_socks_pending(void *data)
{
  impulse_transport_socks server = (impulse_transport_socks)data;
  impulse_int ret;
  fd_set set;
  struct timeval tv = {0,0};
  
  FD_ZERO(&set);
  FD_SET(server->handle, &set);
  
  if ((ret = select(server->handle+1, &set, NULL, NULL, &tv)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot select on sock server socket", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    impulse_unistring_destroy(desc);
  }
  
  if (ret > 0)
    return 1;
  else
    return 0;
}


impulse_int impulse_transport_socks_accept(void *data,
					    void **client)
{
  impulse_transport_socks server = (impulse_transport_socks)data;
  impulse_int handle = -1;
  
  handle = accept(server->handle, NULL, NULL);
  if (handle == -1)
    return -1;

  *client = impulse_transport_sockc_create_handle(handle);
  
  return 0;
}


impulse_int impulse_transport_socks_address(void *data,
					    impulse_unistring *address)
{
  impulse_transport_socks server = (impulse_transport_socks)data;
  
  impulse_transport_sock_join(server->node,
			      server->service,
			      address);
  
  return 0;
}


/* Client functions */


void *impulse_transport_sockc_create(void)
{
  impulse_transport_sockc client = impulse_alloc(sizeof(struct _impulse_transport_sockc));
  
  client->node_local = NULL;
  client->node_remote = NULL;
  client->service_local = NULL;
  client->service_remote = NULL;
  client->handle = -1;
  
  return client;
}


void *impulse_transport_sockc_create_handle(impulse_int handle)
{
  impulse_transport_sockc client = impulse_alloc(sizeof(struct _impulse_transport_sockc));
  union sockaddr_all addr;
  impulse_int addr_len = sizeof(addr);
  
  client->node_local = NULL;
  client->node_remote = NULL;
  client->service_local = NULL;
  client->service_remote = NULL;
  client->handle = handle;

  /* Fix me: don't return null */
  if (getsockname(handle, &addr.a, &addr_len) < 0)
    return NULL;

  impulse_transport_sock_query(&addr,
			       addr_len,
			       &client->node_local,
			       &client->service_local,
			       SOCK_STREAM);

  if (getpeername(handle, &addr.a, &addr_len) < 0)
    return NULL;
  
  impulse_transport_sock_query(&addr,
			       addr_len,
			       &client->node_remote,
			       &client->service_remote,
			       SOCK_STREAM);
  
  return client;
}


void impulse_transport_sockc_destroy(void *data)
{
  impulse_transport_sockc client = (impulse_transport_sockc)data;
  
  if (client->handle != -1)
    impulse_transport_sockc_close(data);

  if (client->node_local)
    impulse_free(client->node_local);
  if (client->node_remote)
    impulse_free(client->node_remote);
  if (client->service_local)
    impulse_free(client->service_local);
  if (client->service_remote)
    impulse_free(client->service_remote);
  
  impulse_free(client);
}


impulse_bool impulse_transport_sockc_is_open(void *data)
{
  impulse_transport_sockc client = (impulse_transport_sockc)data;
  return client->handle != -1;
}


impulse_int impulse_transport_sockc_open(void *data,
					 impulse_unistring address_local,
					 impulse_unistring address_remote)
{
  impulse_transport_sockc client = (impulse_transport_sockc)data;
  union sockaddr_all *addr_local = NULL;
  union sockaddr_all *addr_remote = NULL;
  impulse_int addr_len_local = 0;
  impulse_int addr_len_remote = 0;
  impulse_int reuseaddr = 1;
  
  if (client->handle != -1)
    impulse_transport_sockc_close(data);

  if (address_local)
  {
    if (impulse_transport_sock_lookup(address_local,
				      &client->node_local,
				      &client->service_local,
				      &addr_local,
				      &addr_len_local,
				      SOCK_STREAM) < 0)
      return -1;
  }
  if (impulse_transport_sock_lookup(address_remote,
				    &client->node_remote,
				    &client->service_remote,
				    &addr_remote,
				    &addr_len_remote,
				    SOCK_STREAM) < 0)
    return -1;
  
  if ((client->handle = socket(addr_remote->a.sa_family,
			       SOCK_STREAM, 0)) == -1)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot create client socket", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    /* Criticals never return */
  }

  if (setsockopt(client->handle, SOL_SOCKET, SO_REUSEADDR, (void *)&reuseaddr, sizeof(reuseaddr)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot reuse client socket address", 8);
    impulse_error_syscall(IMPULSE_ERROR_WARNING, desc);
    impulse_unistring_destroy(desc);
  }
  
  if (address_local)
  {
    if (bind(client->handle, &addr_local->a, addr_len_local) < 0)
    {
      impulse_free(addr_local);
      impulse_free(addr_remote);
      close(client->handle);
      client->handle = -1;
      impulse_free(client->node_local);
      impulse_free(client->service_local);
      impulse_free(client->node_remote);
      impulse_free(client->service_remote);
      client->node_local = NULL;
      client->service_local = NULL;
      client->node_remote = NULL;
      client->service_remote = NULL;
      if (errno != EADDRINUSE)
      {
	impulse_unistring desc = impulse_unistring_create_utf("Cannot bind to server address", 8);
	impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
	impulse_unistring_destroy(desc);
      }
      return -1;
    }
  }
  
  if (connect(client->handle, &addr_remote->a, addr_len_remote) < 0)
  {
    impulse_free(addr_local);
    impulse_free(addr_remote);
    close(client->handle);
    client->handle = -1;
    impulse_free(client->node_local);
    impulse_free(client->service_local);
    impulse_free(client->node_remote);
    impulse_free(client->service_remote);
    client->node_local = NULL;
    client->service_local = NULL;
    client->node_remote = NULL;
    client->service_remote = NULL;
    return -1;
  }

  impulse_free(addr_local);
  impulse_free(addr_remote);
  
  return 0;
}


impulse_int impulse_transport_sockc_close(void *data)
{
  impulse_transport_sockc client = (impulse_transport_sockc)data;
  impulse_int ret = 0;
  
  if (client->handle != -1)
    ret = close(client->handle);
  client->handle = -1;
  
  if (ret == -1)
    return -1;
  return 0;
}


impulse_int impulse_transport_sockc_handle(void *data)
{
  impulse_transport_sockc client = (impulse_transport_sockc)data;
  
  return client->handle;
}


impulse_int impulse_transport_sockc_pending(void *data)
{
  impulse_transport_sockc client = (impulse_transport_sockc)data;
  impulse_int pending;
  impulse_int ret;
  
  ret = ioctl(client->handle, FIONREAD, &pending);

  if (ret == -1)
    return -1;
  return pending;
}


impulse_int impulse_transport_sockc_read(void *data,
					      impulse_byte *buffer,
					      impulse_int length)
{
  impulse_transport_sockc client = (impulse_transport_sockc)data;
  impulse_int ret;
  
  ret = recv(client->handle, (char *)buffer, length, 0);
  
  if (ret == -1)
    return -1;

  return ret;
}

impulse_int impulse_transport_sockc_write(void *data,
					       impulse_byte *buffer,
					       impulse_int length)
{
  impulse_transport_sockc client = (impulse_transport_sockc)data;
  impulse_int ret;
  
  ret = send(client->handle, (char *)buffer, length, 0);
  
  if (ret == -1)
    return -1;
  
  return ret;
}


impulse_int impulse_transport_sockc_address_local(void *data,
						       impulse_unistring *address)
{
  impulse_transport_sockc client = (impulse_transport_sockc)data;
  
  impulse_transport_sock_join(client->node_local,
			      client->service_local,
			      address);
  return 0;
}


impulse_int impulse_transport_sockc_address_remote(void *data,
							impulse_unistring *address)
{
  impulse_transport_sockc client = (impulse_transport_sockc)data;
  
  impulse_transport_sock_join(client->node_remote,
			      client->service_remote,
			      address);
  
  return 0;
}


/* Peer functions */

void *impulse_transport_sockp_create(void)
{
  impulse_transport_sockp peer = impulse_alloc(sizeof(struct _impulse_transport_sockp));

  peer->handle = -1;
  peer->node = NULL;
  peer->service = NULL;
  
  return peer;
}


void impulse_transport_sockp_destroy(void *data)
{
  impulse_transport_sockp peer = (impulse_transport_sockp)data;
  
  if (peer->handle)
    impulse_transport_sockp_close(data);
  
  if (peer->node)
    impulse_free(peer->node);
  if (peer->service)
    impulse_free(peer->service);
  impulse_free(peer);
}


impulse_bool impulse_transport_sockp_is_open(void *data)
{
  impulse_transport_sockp peer = (impulse_transport_sockp)data;
  
  return peer->handle != -1;
}


impulse_int impulse_transport_sockp_open(void *data,
					 impulse_unistring address)
{
  impulse_transport_sockp peer = (impulse_transport_sockp)data;
  impulse_uint reuseaddr = 1;
  union sockaddr_all *addr;
  impulse_int addr_len;

  if (peer->handle != -1)
    impulse_transport_socks_close(data);
  
  if (impulse_transport_sock_lookup(address,
				    &peer->node,
				    &peer->service,
				    &addr,
				    &addr_len,
				    SOCK_DGRAM) < 0)
    return -1;

  if ((peer->handle = socket(addr->a.sa_family,
			     SOCK_DGRAM, 0)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot create peer socket", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    /* Criticals never return */
  }

  if (setsockopt(peer->handle, SOL_SOCKET, SO_REUSEADDR, (void *)&reuseaddr, sizeof(reuseaddr)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot reuse peer socket address", 8);
    impulse_error_syscall(IMPULSE_ERROR_WARNING, desc);
    impulse_unistring_destroy(desc);
  }

  if (bind(peer->handle, &addr->a, addr_len) < 0)
  {
    impulse_free(addr);
    close(peer->handle);
    peer->handle = -1;
    impulse_free(peer->node);
    impulse_free(peer->service);
    peer->node = NULL;
    peer->service = NULL;
    if (errno != EADDRINUSE)
    {
      impulse_unistring desc = impulse_unistring_create_utf("Cannot bind to peer address", 8);
      impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
      impulse_unistring_destroy(desc);
    }
    return -1;
  }

  return 0;
}


impulse_int impulse_transport_sockp_close(void *data)
{
  impulse_transport_sockp peer = (impulse_transport_sockp)data;
  
  close(peer->handle);
  peer->handle = -1;
  impulse_free(peer->node);
  impulse_free(peer->service);
  
  peer->node = NULL;
  peer->service = NULL;
  
  return 0;
}


impulse_int impulse_transport_sockp_handle(void *data)
{
  impulse_transport_sockp peer = (impulse_transport_sockp)data;
  
  return peer->handle;
}


impulse_int impulse_transport_sockp_pending(void *data)
{
  impulse_transport_sockp peer = (impulse_transport_sockp)data;
  impulse_int ret;
  fd_set set;
  struct timeval tv = {0,0};
  
  FD_ZERO(&set);
  FD_SET(peer->handle, &set);
  
  if ((ret = select(peer->handle+1, &set, NULL, NULL, &tv)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot select on sock peer socket", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    impulse_unistring_destroy(desc);
  }
  
  if (ret > 0)
    return 1;
  else
    return 0;
}


impulse_int impulse_transport_sockp_send(void *data,
					 impulse_message msg,
					 impulse_unistring address)
{
  impulse_transport_sockp peer = (impulse_transport_sockp)data;
  union sockaddr_all *addr = NULL;
  impulse_int addr_len = sizeof(addr);
  impulse_char *node = NULL;
  impulse_char *service = NULL;
  impulse_int ret = 0;
  
  if (impulse_transport_sock_lookup(address,
				    &node,
				    &service,
				    &addr,
				    &addr_len,
				    SOCK_DGRAM) < 0)
    return -1;

  impulse_free(node);
  
  if ((ret = sendto(peer->handle, msg->data, msg->data_length, 0, &addr->a, addr_len)) < 0)
    return -1;

  return ret;
}


impulse_int impulse_transport_sockp_receive(void *data,
					    impulse_message *msg,
					    impulse_unistring *address)
{
  impulse_transport_sockp peer = (impulse_transport_sockp)data;
  union sockaddr_all *addr = NULL;
  impulse_char *node = NULL;
  impulse_char *service = NULL;
  impulse_int ret = 0;
  impulse_int addr_len = sizeof(addr);
  struct _impulse_message_header header;

  if ((ret = recvfrom(peer->handle, (char *)&header, sizeof(header), MSG_PEEK, NULL, NULL)) < 0)
    return -1;

  *msg = impulse_message_create(0, 0, header.length);
  
  if ((ret = recvfrom(peer->handle, (*msg)->data, (*msg)->data_length, 0, &addr->a, &addr_len)) < 0)
  {
    impulse_message_destroy(*msg);
    *msg = NULL;
    return -1;
  }

  impulse_transport_sock_query(addr,
			       addr_len,
			       &node,
			       &service,
			       SOCK_DGRAM);
  impulse_transport_sock_join(node,
			      service,
			      address);
  impulse_free(node);

  return ret;
}


impulse_int impulse_transport_sockp_address(void *data,
					    impulse_unistring *address)
{
  impulse_transport_sockp peer = (impulse_transport_sockp)data;
  
  impulse_transport_sock_join(peer->node,
			      peer->service,
			      address);
  return 0;
}

#else

impulse_transport impulse_transport_sock = NULL;

#endif

/* End of file */
