/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the ROWS widget library
 * for Impulse.
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * bitmap.c
 * --------
 *
 * Device-independant bitmap widget
 */

#include "../../../config.h"

#include "impulse/rows/bitmap_int.h"
#include "impulse/util/memory.h"
#include "impulse/util/thread.h"
#include "impulse/util/hash.h"
#include "impulse/util/error.h"

#include <stdlib.h>

/*                            ======--------======                           */

impulse_int _p_ROWS_bitmap_ID   = -1;
impulse_int _p_ROWS_bitmap_type = -1;

static void ROWS_bitmap_setup  (ROWS_widget *widget,
				microns width,
				microns height,
				impulse_int coldepth,
				ROWS_bitmap_mask_type  use_mask);
static void ROWS_bitmap_destroy(void **private);

static impulse_hash file_handlers = NULL;
#ifdef USE_THREADS
static impulse_mutex_static file_handlers_lock = IMPULSE_MUTEX_CREATE_STATIC;
#endif

/*                            ======--------======                           */

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_bitmap_create(microns width,
			       microns height,
			       impulse_int coldepth,
			       ROWS_bitmap_mask_type mask)
{
  ROWS_widget *wid;

  if (_p_ROWS_bitmap_ID   == -1)
    _p_ROWS_bitmap_ID   = ROWS_drawable_next_identifier();

  wid = ROWS_bitmap_initialise(_p_ROWS_bitmap_ID,
			       width,
			       height,
			       coldepth,
			       mask);

  return *wid;
}

impulse_int ROWS_bitmap_next_identifier(void)
{
  if (_p_ROWS_bitmap_ID==-1)
    _p_ROWS_bitmap_ID = ROWS_drawable_next_identifier();

  return _p_ROWS_bitmap_ID + 1;
}

ROWS_bitmap ROWS_bitmap_cast(void **widget)
{
  if (_p_ROWS_bitmap_ID   == -1)
    _p_ROWS_bitmap_ID   = ROWS_drawable_next_identifier();
  if (_p_ROWS_bitmap_type == -1)
    _p_ROWS_bitmap_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, _p_ROWS_bitmap_ID, _p_ROWS_bitmap_type);
}

ROWS_widget *ROWS_bitmap_initialise     (impulse_int id,
					 microns width,
					 microns height,
					 impulse_int coldepth,
					 ROWS_bitmap_mask_type mask)
{
  ROWS_widget *widget;

  if (_p_ROWS_bitmap_ID   == -1)
    _p_ROWS_bitmap_ID   = ROWS_drawable_next_identifier();
  if (_p_ROWS_bitmap_type == -1)
    _p_ROWS_bitmap_type = ROWS_widget_next_type();


  widget = ROWS_drawable_initialise(id);
  ROWS_widget_set_type(widget, _p_ROWS_bitmap_ID, _p_ROWS_bitmap_type);
  ROWS_widget_set_destructor(widget, _p_ROWS_bitmap_ID, ROWS_bitmap_destroy);

  ROWS_bitmap_setup(widget,
		    width,
		    height,
		    coldepth,
		    mask);

  return widget;
}

void ROWS_bitmap_setup(ROWS_widget           *widget,
		       microns                w,
		       microns                h,
		       impulse_int            coldepth,
		       ROWS_bitmap_mask_type  mask)
{
  int                   lineskip;
  int                   ncols;
  int                   width, height;
  _p_ROWS_bitmap_data **private;

  switch (coldepth)
    {
    case 15:
      coldepth = 16;
      break;
    case 24:
      coldepth = 32;
      break;
    }
  
  private  = (_p_ROWS_bitmap_data **) ROWS_widget_private(widget, _p_ROWS_bitmap_ID);
  *private = impulse_alloc(sizeof(_p_ROWS_bitmap_data));

  width = MICRONS_TO_PIXEL(w);
  height = MICRONS_TO_PIXEL(h);
  
  (*private)->draw_data = NULL;
  (*private)->format  = coldepth;
  (*private)->width   = width;
  (*private)->height   = height;

  ncols = 1<<coldepth;

  lineskip  = (32 - ((width*coldepth)%32))/coldepth;
  lineskip += width;

  (*private)->lineskip = (lineskip*coldepth)/(8*sizeof(impulse_int));
  (*private)->data     = impulse_alloc((*private)->lineskip*height*sizeof(impulse_int));
  
  if (coldepth<15)
    {
      (*private)->palette = impulse_alloc(ncols*3*sizeof(impulse_ubyte));
    }
  else
    {
      switch (coldepth)
	{
	case 16:
	  /* 5 bits per channel */
	  (*private)->palette = impulse_alloc(32*3*sizeof(impulse_ubyte));
	  break;

	case 32:
	  /* 8 bits per channel */
	  (*private)->palette = impulse_alloc(256*3*sizeof(impulse_ubyte));
	  break;
	}
    }

  switch (mask)
    {
    case ROWS_BITMAP_MASK_NONE:
      (*private)->mask = NULL;
      break;

    case ROWS_BITMAP_MASK_SIMPLE:
      (*private)->mask = impulse_alloc(width*height*sizeof(impulse_ubyte));
      break;

    case ROWS_BITMAP_MASK_ALPHA:
      (*private)->mask = impulse_alloc(width*height*sizeof(impulse_ubyte));
      break;
    }

  (*private)->mask_type = mask;
}

void ROWS_bitmap_destroy(void **data)
{
  _p_ROWS_bitmap_data   **private;
   
  private = (_p_ROWS_bitmap_data **) data;

  impulse_free((*private)->data);
  impulse_free((*private)->palette);

  impulse_free(*private);
}

/*                            ======--------======                           */

void *_p_ROWS_bitmap_get_private_data(ROWS_bitmap              bitmap,
				      ROWS_drawable            drawable,
				      int                      size,
				      _p_ROWS_bitmap_draw_func create,
				      _p_ROWS_bitmap_draw_func destroy)
{
   _p_ROWS_bitmap_data     **private;
   _p_ROWS_bitmap_drawable  *next;
   void                     *found_data = NULL;
  
  ROWS_widget_check_type(bitmap, _p_ROWS_bitmap_type, "ROWS_bitmap");

  private = (_p_ROWS_bitmap_data **) bitmap;

  ROWS_widget_lock(bitmap);

  next = (*private)->draw_data;
  while (next != NULL)
    {
      if (next->drawable == drawable)
	found_data = next->int_data;

      next = next->next;
    }

  if (found_data == NULL)
    {
      next               = impulse_alloc(sizeof(_p_ROWS_bitmap_drawable));

      next->int_data     = impulse_alloc(size);
      next->drawable     = drawable;
      next->destroy_func = destroy;
      next->next         = (*private)->draw_data;

      (*private)->draw_data = next;

      (create)(next->int_data);

      found_data = next->int_data;
    }
  
  ROWS_widget_unlock(bitmap);

  return found_data;
}

/*                            ======--------======                           */

impulse_uint *ROWS_bitmap_get_data(ROWS_bitmap  bitmap)
{
  _p_ROWS_bitmap_data **private;
  
  ROWS_widget_check_type(bitmap, _p_ROWS_bitmap_type, "ROWS_bitmap");

  private = (_p_ROWS_bitmap_data **) bitmap;

  return (*private)->data;
}

ROWS_bitmap ROWS_bitmap_convert(ROWS_bitmap  bitmap,
				impulse_int    newdepth,
				impulse_ubyte *newpalette)
{
  _p_ROWS_bitmap_data **old, **new;
  
  ROWS_widget_check_type(bitmap, _p_ROWS_bitmap_type, "ROWS_bitmap");

  ROWS_widget_lock(bitmap);

  old = (_p_ROWS_bitmap_data **) bitmap;
  new = (_p_ROWS_bitmap_data **)
    ROWS_bitmap_cast(ROWS_bitmap_create(PIXEL_TO_MICRONS((*old)->width),
					PIXEL_TO_MICRONS((*old)->height),
					newdepth,
					(*old)->mask_type));
  
  /* How we proceed depends on the source type... */
  switch ((*old)->format)
    {
    case 1:
    case 2:
    case 4:
    case 8:
      /* Make a map of source pixel to destination pixel colour */
      {
	impulse_int map[256];
	int col;
	int ncols, ncolsnew;

	int x,y, pix,npix,npix2, mask;
	
	ncols = 1<<((*old)->format);
	ncolsnew = 1<<newdepth;

	if (newdepth<15)
	  {
	    for (col=0; col<ncols; col++)
	      {
		int newcol, mindist, col2;
		int r,g,b;

		/* FIXME: there must be a better way of doing this... */

		newcol = -1;
		mindist = 256*256*3;

		r = (*old)->palette[col*3+0];
		g = (*old)->palette[col*3+1];
		b = (*old)->palette[col*3+2];

		for (col2 = 0; col2 < ncolsnew; col2++)
		  {
		    int dist;
		    int rdist, gdist, bdist;

		    rdist = (r-newpalette[col2*3+0]);
		    gdist = (g-newpalette[col2*3+1]);
		    bdist = (b-newpalette[col2*3+2]);

		    dist = rdist*rdist + gdist*gdist + bdist*bdist;

		    if (dist<mindist)
		      {
			mindist = dist;
			newcol  = col2;
		      }
		  }

		map[col] = newcol;
	      }
	  }
	else
	  {
	    switch (newdepth)
	      {
	      case 32:
		/* Heh - easy. Note gamma correction effects  */
		for (col=0; col<ncols; col++)
		  {
		    int r, g, b;
		    
		    r = (*old)->palette[col*3+0];
		    g = (*old)->palette[col*3+1];
		    b = (*old)->palette[col*3+2];

		    map[col] = r + (g<<8) + (b<<16);
		  }
		break;

	      case 16:
		/* Just as easy. Also note gamma correction effects */
		for (col=0; col<ncols; col++)
		  {
		    int r, g, b;
		    
		    r = (*old)->palette[col*3+0] >> 3;
		    g = (*old)->palette[col*3+1] >> 3;
		    b = (*old)->palette[col*3+2] >> 3;

		    map[col] = r + (g<<5) + (b<<10);
		  }
		break;
	      }
	  }

	npix  = 32/(*old)->format;
	npix2 = 32/(*new)->format;
	mask  = (1<<(*old)->format)-1;

	/* FIXME: optimise this */

	for (y=0; y<(*old)->height; y++)
	  {
	    int nx = 0;
	    for (x=0; x<(*old)->lineskip; x++)
	      {
		impulse_uint oldword;

		oldword = (*old)->data[nx+(y*(*old)->lineskip)];
		  
		for (pix=0; pix<npix; pix++)
		  {
 		    int oldpix;

		    oldpix = oldword&mask;

		    if ((nx%npix2) == 0)
		      (*new)->data[(nx/npix2)+(y*(*new)->lineskip)] =
			map[oldpix];
		    else
		      (*new)->data[(nx/npix2)+(y*(*new)->lineskip)] |=
			map[oldpix]<<((nx%npix2)*newdepth);
		    
		    oldword >>= (*old)->format;

		    nx++;

		    if (nx >= (*new)->width)
		      pix = npix; /* FIXME: Slow, usually redundant */
		  }
	      }
	  }

	switch (newdepth)
	  {
	  case 16:
	    memcpy((*new)->palette,
		   newpalette,
		   32*3*sizeof(impulse_ubyte));
	    break;

	  case 32:
	    memcpy((*new)->palette,
		   newpalette,
		   256*3*sizeof(impulse_ubyte));
	    break;

	  default:
	    memcpy((*new)->palette,
		   newpalette,
		   (1<<newdepth)*3*sizeof(impulse_ubyte));
	    break;
	  }
      }
      break;

    case 16:
    case 32:
      switch (newdepth)
	{
	case 16:
	  /* Translating between truecolour types is easy... */
	  if ((*old)->format == 16)
	    {
	      memcpy((*new)->data,
		     (*old)->data,
		     (*old)->lineskip*(*old)->height*sizeof(impulse_uint));
	    }
	  else
	    {
	      int x,y;

	      for (y=0; y<(*old)->height; y++)
		{
		  for (x=0; x<(*old)->lineskip; x++)
		    {
		      int r,g,b;
		      impulse_uint oldword, newword;

		      oldword = (*old)->data[x+(y*(*old)->lineskip)];

		      r =  oldword      &255;
		      g = (oldword>>8)  &255;
		      b = (oldword>>16) &255;

		      r >>= 3;
		      g >>= 3;
		      b >>= 3;

		      newword = r | (g<<5) | (b<<10);

		      if ((x+1)<(*old)->lineskip)
			{
			  x++;

			  oldword = (*old)->data[x+(y*(*old)->lineskip)];

			  r =  oldword      &255;
			  g = (oldword>>8)  &255;
			  b = (oldword>>16) &255;
			  
			  r >>= 3;
			  g >>= 3;
			  b >>= 3;
			  
			  newword |= r<<16 | (g<<21) | (b<<26);
			}

		      (*new)->data[(x/2)+(y*(*new)->lineskip)] = newword;
		    }
		}
	    }
	  break;

	case 32:
	  /* Translating between truecolour types is easy... */
	  if ((*old)->format == 32)
	    {
	      memcpy((*new)->data,
		     (*old)->data,
		     (*old)->lineskip*(*old)->height*sizeof(impulse_uint));
	    }
	  break;

	default:
	  {
	    impulse_unistring desc = impulse_unistring_create_utf("Cannot convert bitmap format", 8);
	    impulse_error_raise(IMPULSE_ERROR_WARNING, desc, NULL);
	    impulse_unistring_destroy(desc);
	  }
	  break;
	}
      break;
    }

  ROWS_widget_unlock(bitmap);

  return (ROWS_bitmap) new;
}

/*                            ======--------======                           */


static _p_ROWS_bitmap_handler load_handler(impulse_unistring type)
{
  impulse_module module = NULL;
  _p_ROWS_bitmap_handler _p_handler = NULL;
  ROWS_bitmap_handler handler = NULL;
  impulse_char *envpath = getenv("IMPULSE_BITMAP_PATH");
  impulse_unistring path = NULL;
  impulse_unistring sym = NULL;

  if (envpath)
  {
    path = impulse_unistring_create_utf(envpath, 8);
    impulse_unistring_append_utf(path, "/", 8);
  }
  else
  {
    path = impulse_unistring_create_utf(PREFIX_DIR, 8);
    impulse_unistring_append_utf(path, "/lib/impulse/modules/bitmap/", 8);
  }
  
  impulse_unistring_append(path, type);
    
  if (!(module = impulse_module_create(path)))
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot load image file handler", 8);
    impulse_error_raise(IMPULSE_ERROR_WARNING, desc, path);
    impulse_unistring_destroy(desc);
    impulse_unistring_destroy(path);
    return NULL;
  }
  
  sym = impulse_unistring_create_utf("handler", 8);
  handler = (ROWS_bitmap_handler)impulse_module_symbol(module, sym);
  impulse_unistring_destroy(sym);
  
  if (!handler)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Incomplete image file handler", 8);
    impulse_error_raise(IMPULSE_ERROR_WARNING, desc, path);
    impulse_unistring_destroy(desc);
    impulse_unistring_destroy(path);
    return NULL;
  }  
  impulse_unistring_destroy(path);

  _p_handler = impulse_alloc(sizeof(struct __p_ROWS_bitmap_handler));
  _p_handler->handler = handler;
  _p_handler->module = module;
  
  return _p_handler;
}


static void unload_handler(_p_ROWS_bitmap_handler _p_handler)
{
  impulse_module_destroy(_p_handler->module);
  impulse_free(_p_handler);
}


static ROWS_bitmap_handler find_handler(impulse_unistring filename)
{
  impulse_int offset = impulse_unistring_rfind_chr(filename, (impulse_unichar)'.');
  impulse_unistring type;
  ROWS_bitmap_handler handler = NULL;
  _p_ROWS_bitmap_handler _p_handler = NULL;
  
  if (offset < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot determine image file type", 8);
    impulse_error_raise(IMPULSE_ERROR_WARNING, desc, filename);
    impulse_unistring_destroy(desc);
    return NULL;
  }

  type = impulse_unistring_create_empty();
  impulse_unistring_substring(type, filename, offset+1, filename->length);

#if USE_THREADS
  impulse_mutex_lock(&file_handlers_lock);
#endif
  
  if (!file_handlers)
  {
    file_handlers = impulse_hash_create();
    impulse_hash_cleaner(file_handlers, (impulse_hash_cleaner_func)unload_handler);
  }
  
  if (!(_p_handler = impulse_hash_recover(file_handlers, type)))
    _p_handler = load_handler(type);
  
  if (_p_handler)
    handler = _p_handler->handler;
  impulse_unistring_destroy(type);
  
#if USE_THREADS
  impulse_mutex_unlock(&file_handlers_lock);
#endif
  
  return handler;
}


void ROWS_bitmap_load(ROWS_bitmap bitmap,
		      impulse_unistring filename)
{
  ROWS_bitmap_handler handler = find_handler(filename);
  
  if (!handler)
    return;
  
  if (!(handler->load)(bitmap, filename))
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot load image file", 8);
    impulse_error_raise(IMPULSE_ERROR_WARNING, desc, filename);
    impulse_unistring_destroy(desc);
  }
}


void ROWS_bitmap_save(ROWS_bitmap bitmap,
		      impulse_unistring filename)
{
  ROWS_bitmap_handler handler = find_handler(filename);
  
  if (!handler)
    return;
  
  if (!(handler->save)(bitmap, filename))
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot save image file", 8);
    impulse_error_raise(IMPULSE_ERROR_WARNING, desc, filename);
    impulse_unistring_destroy(desc);
  }
}


/* End of file */
