/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * box.c
 * -----
 *
 * Base widget class
 */

#include "../../../config.h"

#include "impulse/rows/box_int.h"
#include "impulse/util/memory.h"
#include "impulse/util/error.h"

/*                            ======--------======                           */

static void ROWS_box_setup(ROWS_widget *widget);
static int  box_ID   = -1;
static int  box_type = -1;

static void ROWS_box_destroy(void **private);

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_box_create(void)
{
  ROWS_widget *wid;
  impulse_unistring desc = impulse_unistring_create_utf("Box widget created at top level (did you really mean this?)", 8);
  
  impulse_error_raise(IMPULSE_ERROR_WARNING, desc, NULL);
  impulse_unistring_destroy(desc);

  if (box_ID   == -1)
    box_ID   = ROWS_widget_next_identifier();

  wid = ROWS_box_initialise(box_ID);

  return *wid;
}

ROWS_widget *ROWS_box_initialise(int id)
{
  ROWS_widget *widget;

  if (box_ID   == -1)
    box_ID   = ROWS_widget_next_identifier();
  if (box_type == -1)
    box_type = ROWS_widget_next_type();

  widget = ROWS_widget_initialise(id);
  ROWS_widget_set_type(widget, box_ID, box_type);
  ROWS_widget_set_destructor(widget, box_ID, ROWS_box_destroy);

  ROWS_box_setup(widget);

  return widget;
}

int ROWS_box_next_identifier(void)
{
  if (box_ID==-1)
    box_ID = ROWS_widget_next_identifier();

  return box_ID + 1;
}

ROWS_box ROWS_box_cast(void **widget)
{
  if (box_ID   == -1)
    box_ID   = ROWS_widget_next_identifier();
  if (box_type == -1)
    box_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, box_ID, box_type);
}

void ROWS_box_setup(ROWS_widget *widget)
{
  _p_ROWS_box_data **private;

  private  = (_p_ROWS_box_data **) ROWS_widget_private(widget, box_ID);
  *private = impulse_alloc(sizeof(_p_ROWS_box_data));

  (*private)->virtual_box.x      = -1;
  (*private)->virtual_box.y      = -1;

  (*private)->physical_box       = (*private)->virtual_box;

  (*private)->scroll.x           = 0;
  (*private)->scroll.y           = 0;
}

void ROWS_box_destroy(void **private)
{
  impulse_free(*private);
}

/*                            ======--------======                           */

/* Class-specific functions */

void ROWS_box_set_position(ROWS_box    box,
			   ROWS_point *position)
{
  _p_ROWS_box_data **private;

  ROWS_widget_lock(box);
  ROWS_widget_check_type(box, box_type, "ROWS_box");

  private  = (_p_ROWS_box_data **) box;

  (*private)->position = *position;
  ROWS_widget_unlock(box);
}

ROWS_point *ROWS_box_get_position(ROWS_box box)
{
  _p_ROWS_box_data **private;

  ROWS_widget_check_type(box, box_type, "ROWS_box");

  private  = (_p_ROWS_box_data **) box;

  return &(*private)->position;
}

void ROWS_box_set_boundaries(ROWS_box    box,
			     ROWS_point *boundaries)
{
  _p_ROWS_box_data **private;

  ROWS_widget_lock(box);
  ROWS_widget_check_type(box, box_type, "ROWS_box");

  private  = (_p_ROWS_box_data **) box;

  (*private)->virtual_box = (*private)->physical_box = *boundaries;
  ROWS_widget_unlock(box);
}

void ROWS_box_get_boundaries(ROWS_box    box,
			     ROWS_rect *rect)
{
  _p_ROWS_box_data **private;

  ROWS_widget_check_type(box, box_type, "ROWS_box");

  private  = (_p_ROWS_box_data **) box;

  rect->pos  = (*private)->position;
  rect->size = (*private)->physical_box;
}

void ROWS_box_set_physical(ROWS_box    box,
			   ROWS_point *physical)
{
  _p_ROWS_box_data **private;

  ROWS_widget_lock(box);
  ROWS_widget_check_type(box, box_type, "ROWS_box");

  private  = (_p_ROWS_box_data **) box;

  (*private)->physical_box = *physical;
  ROWS_widget_unlock(box);
}

void ROWS_box_set_virtual(ROWS_box    box,
			  ROWS_point *virtual)
{
  _p_ROWS_box_data **private;

  ROWS_widget_lock(box);
  ROWS_widget_check_type(box, box_type, "ROWS_box");

  private  = (_p_ROWS_box_data **) box;

  (*private)->virtual_box = *virtual;
  ROWS_widget_unlock(box);
}

ROWS_point *ROWS_box_get_virtual(ROWS_box box)
{
  _p_ROWS_box_data **private;

  ROWS_widget_check_type(box, box_type, "ROWS_box");

  private  = (_p_ROWS_box_data **) box;

  return &(*private)->virtual_box;
}

ROWS_point *ROWS_box_get_physical(ROWS_box box)
{
  _p_ROWS_box_data **private;

  ROWS_widget_check_type(box, box_type, "ROWS_box");

  private  = (_p_ROWS_box_data **) box;

  return &(*private)->physical_box;
}

/* FIXME: check types here */
void ROWS_box_set_scroll(ROWS_box    box,
			 ROWS_point *scroll)
{
  _p_ROWS_box_data **private;

  ROWS_widget_lock(box);
  ROWS_widget_check_type(box, box_type, "ROWS_box");
  private  = (_p_ROWS_box_data **) box;

  (*private)->scroll = *scroll;
  ROWS_widget_unlock(box);
}

ROWS_point *ROWS_box_get_scroll(ROWS_box box)
{
  _p_ROWS_box_data **private;

  ROWS_widget_check_type(box, box_type, "ROWS_box");

  private  = (_p_ROWS_box_data **) box;

  return &(*private)->scroll;
}
