/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * child.c
 * -------
 *
 * Class of widgets child in containers
 */

#include "../../../config.h"

#include "impulse/rows/child_int.h"

#include "impulse/util/memory.h"
#include "impulse/util/unistring.h"
#include "impulse/util/sparse.h"
#include "impulse/util/thread.h"

#if HAVE_GUI_XWIN
# include "impulse/rows/child_xwin.c"
#else
#error "Child windows not implemented on this platform"
#endif

/*                            ======--------======                           */

static void ROWS_child_setup(ROWS_widget   *widget,
			     ROWS_drawable  drawable);
       int  _p_ROWS_child_ID   = -1;
       int  _p_ROWS_child_type = -1;

static void ROWS_child_destroy(void **private);

static int _p_ROWS_child_redraw(ROWS_child            widget,
				ROWS_child_event_num  event_number,
				ROWS_child_event     *event,
				void                 *data);
static int _p_ROWS_child_resize(ROWS_child            widget,
				ROWS_child_event_num  event_number,
				ROWS_child_event     *event,
				void                 *data);

/*                            ======--------======                           */

static impulse_sparse selected_children = NULL;
typedef struct _p_ROWS_child_select
{
  ROWS_child select;
} _p_ROWS_child_select;
#if USE_THREADS
static impulse_mutex set_state_mutex = NULL;
#endif

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_child_create(ROWS_drawable drawable)
{
  ROWS_widget *wid;

  if (_p_ROWS_child_ID   == -1)
    _p_ROWS_child_ID   = ROWS_eventbox_next_identifier();

  wid = ROWS_child_init_drawable(_p_ROWS_child_ID,
				 drawable);

  return *wid;
}

ROWS_widget *ROWS_child_init_drawable(int id,
				      ROWS_drawable drawable)
{
  ROWS_widget *widget;

  if (_p_ROWS_child_ID   == -1)
    _p_ROWS_child_ID   = ROWS_eventbox_next_identifier();
  if (_p_ROWS_child_type == -1)
    _p_ROWS_child_type = ROWS_widget_next_type();

  widget = ROWS_eventbox_initialise(id);
  ROWS_widget_set_type(widget, _p_ROWS_child_ID, _p_ROWS_child_type);
  ROWS_widget_set_destructor(widget, _p_ROWS_child_ID, ROWS_child_destroy);

  ROWS_child_setup(widget,
		   drawable);

  return widget;
}

ROWS_widget *ROWS_child_initialise(int id)
{
  ROWS_widget *widget;

  if (_p_ROWS_child_ID   == -1)
    _p_ROWS_child_ID   = ROWS_eventbox_next_identifier();
  if (_p_ROWS_child_type == -1)
    _p_ROWS_child_type = ROWS_widget_next_type();

  widget = ROWS_eventbox_initialise(id);
  ROWS_widget_set_type(widget, _p_ROWS_child_ID, _p_ROWS_child_type);
  ROWS_widget_set_destructor(widget, _p_ROWS_child_ID, ROWS_child_destroy);

  ROWS_child_setup(widget,
		   NULL);

  return widget;
}

int ROWS_child_next_identifier(void)
{
  if (_p_ROWS_child_ID==-1)
    _p_ROWS_child_ID = ROWS_eventbox_next_identifier();

  return _p_ROWS_child_ID + 1;
}

ROWS_child ROWS_child_cast(void **widget)
{
  if (_p_ROWS_child_ID   == -1)
    _p_ROWS_child_ID   = ROWS_eventbox_next_identifier();
  if (_p_ROWS_child_type == -1)
    _p_ROWS_child_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, _p_ROWS_child_ID, _p_ROWS_child_type);
}

void ROWS_child_setup(ROWS_widget   *widget,
		      ROWS_drawable  drawable)
{
  _p_ROWS_child_data **private;

  private  = (_p_ROWS_child_data **) ROWS_widget_private(widget, _p_ROWS_child_ID);
  *private = impulse_alloc(sizeof(_p_ROWS_child_data));

  if (!selected_children)
  {
    impulse_mutex_global_lock;
    if (!selected_children)
    {
#if USE_THREADS
      set_state_mutex = impulse_mutex_create_recursive();
#endif
      selected_children = impulse_sparse_create();
    }
    impulse_mutex_global_unlock;
  }

  ROWS_eventbox_register_widget(ROWS_eventbox_cast(*widget),
				_p_ROWS_child_ID,
				ROWS_CHILD_MAX_EVENT);

  (*private)->draw = drawable;

  (*private)->active   = ROWS_STYLE_NORMAL;
  (*private)->border.x = 1;
  (*private)->border.y = 1;
  (*private)->origin.x = 0;
  (*private)->origin.y = 0;

  (*private)->realised   = 0;
  (*private)->do_redraw  = 1;
  (*private)->activation = ROWS_STYLE_NORMAL;

  (*private)->container = NULL;

  (*private)->state        = ROWS_CHILD_UNSELECTED;
  (*private)->select_group = 0;
  
  if (drawable!=NULL)
    {
      (*private)->style = ROWS_drawable_get_style(drawable);
      ROWS_style_reference((*private)->style);
    }
  else
    (*private)->style = NULL;
      
  ROWS_child_register_event(ROWS_child_cast(*widget),
			    ROWS_CHILD_REDRAW,
			    _p_ROWS_child_redraw,
			    NULL);
  ROWS_child_register_event(ROWS_child_cast(*widget),
			    ROWS_CHILD_RESIZE,
			    _p_ROWS_child_resize,
			    NULL);

  _p_ROWS_child_int_setup(widget, drawable);
}

void ROWS_child_destroy(void **data)
{
  _p_ROWS_child_data **private;

  private  = (_p_ROWS_child_data **) data;

  if ((*private)->select_group != 0)
    {
      /*
       * The observant among you will note that set_state_mutex must
       * be initialised for this widget to have a grouping
       */
#if USE_THREADS
      impulse_mutex_lock(set_state_mutex);
#endif
      if ((*private)->select_group != 0)
	{
	  impulse_sparse_add(selected_children,
			     (*private)->select_group,
			     NULL);
	}
#if USE_THREADS
      impulse_mutex_unlock(set_state_mutex);
#endif
    }
  
  _p_ROWS_child_int_destroy(private);
  ROWS_style_dereference((*private)->style);

  impulse_free(*private);
}

/*                            ======--------======                           */

/* Class-specific functions */

/* Locking functions */

void ROWS_child_lock(ROWS_child child)
{
#if USE_THREADS
  _p_ROWS_child_data **private;
  ROWS_container       lock_container;
  
  if (child == NULL)
    return;

  private = (_p_ROWS_child_data **) ROWS_child_cast(child);

  /* We lock the container first, then the child */
  do {
    lock_container = (*private)->container;
    ROWS_child_lock(lock_container);
    (*private)->lock_container = lock_container;
    ROWS_widget_lock(child);

    /* If we've been moved around since we got the lock, deal with it */
    if (lock_container != (*private)->container)
      {
	ROWS_child_unlock(lock_container);
	ROWS_widget_unlock(child);
      }
  } while (lock_container != (*private)->container);
#endif
}

void ROWS_child_unlock(ROWS_child child)
{
#if USE_THREADS
  _p_ROWS_child_data **private;
  
  if (child == NULL)
    return;

  private = (_p_ROWS_child_data **) ROWS_child_cast(child);

  ROWS_child_unlock((*private)->lock_container);
  ROWS_widget_unlock(child);
#endif
}

/* Event functions */

void ROWS_child_register_event(ROWS_child            child,
			       ROWS_child_event_num  event,
			       ROWS_child_func       func,
			       void                 *data)
{
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  ROWS_eventbox_register_handler(ROWS_eventbox_cast(child),
				 _p_ROWS_child_ID,
				 (ROWS_eventbox_func) func,
				 event,
				 data);
}

void ROWS_child_event_send(ROWS_child            child,
			   ROWS_child_event_num  event,
			   ROWS_child_event     *data)
{
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  ROWS_child_lock(child);
  
  ROWS_eventbox_send(ROWS_eventbox_cast(child),
		     _p_ROWS_child_ID,
		     child,
		     event,
		     data);

  if (event == ROWS_CHILD_RESIZE)
    _p_ROWS_child_int_resize(child);
  
  ROWS_child_unlock(child);
}

/* Geometry functions */

void ROWS_child_set_size(ROWS_child  child,
			 ROWS_point *size)
{
  ROWS_child_event event;
  
  ROWS_box_set_physical(ROWS_box_cast(child), size);
  event.event.set_size.newsize = *size;
  ROWS_child_event_send(child,
			ROWS_CHILD_SET_SIZE,
			&event);
}

ROWS_point *ROWS_child_get_size(ROWS_child child)
{
  return ROWS_box_get_physical(ROWS_box_cast(child));
}

void ROWS_child_set_border(ROWS_child  child,
			   ROWS_point *border)
{
  _p_ROWS_child_data **private;
  
  ROWS_widget_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  (*private)->border = *border;
  ROWS_widget_unlock(child);
}

ROWS_point *ROWS_child_get_border(ROWS_child child)
{
  _p_ROWS_child_data **private;
  
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  return &((*private)->border);
}

void ROWS_child_set_origin(ROWS_child  child,
			   ROWS_point *origin)
{
  _p_ROWS_child_data **private;
  
  ROWS_widget_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  (*private)->origin = *origin;
  ROWS_widget_unlock(child);
}

ROWS_point *ROWS_child_get_origin(ROWS_child child)
{
  _p_ROWS_child_data **private;

  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  return &((*private)->origin);  
}

void ROWS_child_container(ROWS_child     child,
			  ROWS_container container)
{
  _p_ROWS_child_int_container(child, container);
}

/* Display/hiding functions */

void ROWS_child_show(ROWS_child child)
{
  _p_ROWS_child_data **private;
  ROWS_child_event     event;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  /*
  if ((*private)->container == NULL)
    _p_ROWS_critical(CHILD_NOCONTAINER, 0, child);
  */

  if ((*private)->realised == 0)
    ROWS_child_realise(child);

  (*private)->shown = 1;

  _p_ROWS_child_int_show(child);

  ROWS_child_event_send(child,
			ROWS_CHILD_SHOW,
			&event);

  if ((*private)->container != NULL)
    {
      ROWS_container_changed((*private)->container,
			     child);
      ROWS_child_update(child);
    }

  ROWS_child_unlock(child);
}

void ROWS_child_hide(ROWS_child child)
{
  _p_ROWS_child_data **private;
  ROWS_child_event     event;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  /*
  if ((*private)->container == NULL)
    _p_ROWS_critical(CHILD_NOCONTAINER, 0, child);
  */

  if ((*private)->realised == 0)
    ROWS_child_realise(child);

  (*private)->shown = 0;

  _p_ROWS_child_int_hide(child);

  ROWS_child_event_send(child,
			    ROWS_CHILD_HIDE,
			    &event);

  if ((*private)->container != NULL)
    {
      ROWS_container_changed((*private)->container,
			     child);
      ROWS_child_update(ROWS_child_cast((*private)->container));
    }

  ROWS_child_unlock(child);
}

/* Activation state functions */

void ROWS_child_set_activation(ROWS_child             child,
			       ROWS_style_active_type type)
{
  _p_ROWS_child_data **private;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  (*private)->active = type;
  ROWS_child_update(child);
  
  ROWS_child_unlock(child);
}

ROWS_style_active_type ROWS_child_get_activation(ROWS_child child)
{
  _p_ROWS_child_data     **private;
  ROWS_style_active_type   style;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  style = (*private)->active;
  
  ROWS_child_unlock(child);

  return style;
}

/* Realisation functions */

void ROWS_child_realise(ROWS_child child)
{
  _p_ROWS_child_data **private;
  ROWS_child_event     event;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  if ((*private)->realised != 0)
    {
      ROWS_child_unlock(child);
      return;
    }
  
  (*private)->realised = 1;

  ROWS_child_event_send(child,
			    ROWS_CHILD_REALISE,
			    &event);

  if ((*private)->container != NULL)
    ROWS_container_size((*private)->container, child);

  _p_ROWS_child_int_realised(child);

  ROWS_child_unlock(child);
}

void ROWS_child_unrealise(ROWS_child child)
{
  _p_ROWS_child_data **private;
  ROWS_child_event     event;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  if ((*private)->realised != 0)
    {
      ROWS_child_unlock(child);
      return;
    }

  ROWS_child_event_send(child,
			    ROWS_CHILD_UNREALISE,
			    &event);
  
  (*private)->realised = 0;
  ROWS_child_unlock(child);
}

/* Information functions */

ROWS_style ROWS_child_get_style(ROWS_child child)
{
  _p_ROWS_child_data **private;
  
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  return (*private)->style;
}

ROWS_drawable ROWS_child_get_drawable(ROWS_child child)
{
  _p_ROWS_child_data **private;
  
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  return (*private)->draw;
}

/* Event functions */

static int _p_ROWS_child_resize(ROWS_child            widget,
				ROWS_child_event_num  event_number,
				ROWS_child_event     *event,
				void                 *data)
{
  ROWS_box_set_virtual(ROWS_box_cast(widget),
		       ROWS_box_get_physical(ROWS_box_cast(widget)));
  
  return 0;
}
     
static int _p_ROWS_child_redraw(ROWS_child            widget,
				ROWS_child_event_num  event_number,
				ROWS_child_event     *event,
				void                 *data)
{
  _p_ROWS_child_data **private;
  ROWS_drawable        drawable;
  ROWS_style          style;
  ROWS_rect            box;
  impulse_unistring    foo;
  ROWS_brush_measure   fontsize;
  
  /* ROWS_child_lock(widget); */

  drawable = ROWS_child_get_drawable(widget);
  private  = (_p_ROWS_child_data **) widget;

  style = ROWS_child_get_style(widget);
  ROWS_box_get_boundaries(ROWS_box_cast(widget),
			  &box);

  /* FIXME: something to internationalise */
  foo = impulse_unistring_create_utf("No widget", 8);
  ROWS_brush_font_measure(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_FOREGROUND],
			  foo,
			  &fontsize);
  ROWS_brush_font_paint(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_FOREGROUND],
			foo,
			(box.size.x-fontsize.width)/2,
			(box.size.y+fontsize.ascend-fontsize.descend)/2);
  impulse_unistring_destroy(foo);
  
  /*
  ROWS_brush_draw_line(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_FOREGROUND],
		       box.pos.x, box.pos.y,
		       box.pos.x+box.size.x, box.pos.y+box.size.y);
  ROWS_brush_draw_line(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_FOREGROUND],
		       box.pos.x+box.size.x, box.pos.y,
		       box.pos.x, box.pos.y+box.size.y);
  ROWS_brush_fill_rectangle(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_BORDERSHADOW],
			    box.pos.x+12, box.pos.y+12,
			    box.size.x-24, box.size.y-24);
  ROWS_brush_fill_arc(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_BORDERHILITE],
		      box.pos.x+24, box.pos.y+24,
		      box.size.x-48, box.size.y-48,
		      24*64, 123*64);

  ROWS_brush_fill_arc(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_FOREGROUND],
		      box.pos.x+40, box.pos.y+40,
		      box.size.x-80, box.size.y-80,
		      0, 360*64);
  */
		      
  /* ROWS_child_unlock(widget); */
  
  return 1; /* Kilroy was here */
}

/* Set-up ready to draw functions */

void ROWS_child_register_drawable(ROWS_child child,
				      ROWS_drawable  drawable)
{
  _p_ROWS_child_data **private;
  
  ROWS_widget_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;
  
  if ((*private)->draw != NULL)
    {
      ROWS_style_dereference((*private)->style);
      
      (*private)->style = NULL;
      (*private)->draw  = NULL;
    }

  if (drawable != NULL)
    {
      (*private)->style = ROWS_drawable_get_style(drawable);
      (*private)->draw  = drawable;
      
      ROWS_style_reference((*private)->style);
    }

  ROWS_widget_unlock(child);
}

extern void ROWS_child_update(ROWS_child child)
{
  _p_ROWS_child_data **private;
  ROWS_rect            rect;

  ROWS_child_lock(child);

  private = (_p_ROWS_child_data **) child;

  if ((*private)->draw == NULL)
    {
      ROWS_child_unlock(child);
      return;
    }

  ROWS_widget_lock((*private)->draw);

  _p_ROWS_child_int_update(child);

  ROWS_box_get_boundaries(ROWS_box_cast(child), &rect);
  ROWS_drawable_clip_region((*private)->draw, 0);
  ROWS_drawable_set_clip   ((*private)->draw, &rect);

  ROWS_drawable_clear_area((*private)->draw, 
			   rect.pos.x, rect.pos.y,
			   rect.size.x, rect.size.y);

  ROWS_drawable_sync((*private)->draw);

  ROWS_widget_unlock((*private)->draw);

  ROWS_child_unlock(child);
}

/*                            ======--------======                           */

void ROWS_child_set_state(ROWS_child       child,
			  ROWS_child_state state,
			  impulse_int            grouping)
{
  _p_ROWS_child_data **private;

  ROWS_widget_lock(child);
#if USE_THREADS
  impulse_mutex_lock(set_state_mutex);
#endif
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  private = (_p_ROWS_child_data **) child;

  if (grouping != 0)
    {
      _p_ROWS_child_select *old_select, *new_select;

      old_select = impulse_sparse_recover(selected_children,
					  grouping);
      if (old_select != NULL)
	if (old_select->select != child && old_select->select != NULL)
	  {
	    ROWS_widget_unlock(child);

	    ROWS_child_set_state(old_select->select,
				 ROWS_CHILD_UNSELECTED,
				 0);
	    
	    ROWS_widget_lock(child);
	  }

      new_select = impulse_alloc(sizeof(_p_ROWS_child_select));
      new_select->select = child;

      impulse_sparse_add(selected_children,
			 grouping,
			 new_select);
    }

  (*private)->state        = state;
  (*private)->select_group = grouping;
#if USE_THREADS
  impulse_mutex_unlock(set_state_mutex);
#endif
  ROWS_widget_unlock(child);  
}
