/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * child.h
 * -------
 *
 * Something contained by a container
 */

#ifndef __impulse_rows_child_h
#define __impulse_rows_child_h

#include <impulse/rows/eventbox.h>
#include <impulse/rows/drawable.h>
#include <impulse/rows/brush.h>
#include <impulse/rows/input.h>

/*                            ======--------======                           */

/* Contained widget type */

typedef void **ROWS_child;

/*                            ======--------======                           */

/* Events */

typedef enum {
  ROWS_CHILD_REALISE    = 0,
  ROWS_CHILD_UNREALISE,
  ROWS_CHILD_REDRAW,
  ROWS_CHILD_RESIZE,
  ROWS_CHILD_POINTER_ENTER,
  ROWS_CHILD_POINTER_LEAVE,
  ROWS_CHILD_POINTER_MOVE,
  ROWS_CHILD_SET_SIZE,
  ROWS_CHILD_KEY,
  ROWS_CHILD_CLICK,
  ROWS_CHILD_MOTION,
  ROWS_CHILD_CHANGE_STATE,
  ROWS_CHILD_SHOW,
  ROWS_CHILD_HIDE,
  ROWS_CHILD_SIZE,

  ROWS_CHILD_MAX_EVENT
} ROWS_child_event_num;

struct ROWS_event_realise {
  int null;
};

struct ROWS_event_unrealise {
  int null;
};

typedef enum _ROWS_child_state
{
  ROWS_CHILD_FOCUSED,
  ROWS_CHILD_UNFOCUSED,
  ROWS_CHILD_SELECTED,
  ROWS_CHILD_UNSELECTED,
  ROWS_CHILD_SHADED
} ROWS_child_state;

typedef enum _ROWS_child_grouping
{
  ROWS_CHILD_PRESSED    = 0x10000,
  ROWS_CHILD_INPUTFOCUS
} ROWS_child_grouping;

struct ROWS_event_state {
  ROWS_child_state new_state;
  impulse_int            grouping;
  
  ROWS_widget      new_focus;
};

struct ROWS_event_visibility {
  int null;
};

struct ROWS_event_size {
  int null;
};

struct ROWS_event_pointer {
  int x, y;
  int state;
};

struct ROWS_event_set_size {
  ROWS_point newsize;
};

struct _ROWS_child_event {
  union {
    struct ROWS_event_realise    realise;
    struct ROWS_event_unrealise  unrealise;
    struct ROWS_event_redraw     redraw;
    struct ROWS_event_resize     resize;
    struct ROWS_event_set_size   set_size;
    struct ROWS_event_pointer    enter;
    struct ROWS_event_pointer    leave;
    struct ROWS_event_pointer    move;
    struct ROWS_event_key        key;
    struct ROWS_event_mouseclick click;
    struct ROWS_event_state      change_state;
    struct ROWS_event_visibility show;
    struct ROWS_event_visibility hide;
    struct ROWS_event_size       size;
  } event;
};

typedef struct _ROWS_child_event ROWS_child_event;

/* Functions */

typedef int (*ROWS_child_func) (ROWS_child            widget,
				ROWS_child_event_num  event_number,
				ROWS_child_event     *event,
				void                     *data);

/*                            ======--------======                           */

/* Standard widget functions */

extern ROWS_widget       ROWS_child_create         (ROWS_drawable drawable);
extern ROWS_widget      *ROWS_child_init_drawable  (int id,
						      ROWS_drawable drawable);
extern ROWS_widget      *ROWS_child_initialise     (int id);
extern inline int        ROWS_child_next_identifier(void);
extern inline ROWS_child ROWS_child_cast           (void **widget);

/*                            ======--------======                           */

/* Class-specific functions */

/* 
 * Note: to prevent some very tricky deadlock situations, use this function
 *       to lock child widgets, NOT ROWS_widget_lock. (Well, by tricky, I
 *       mean it took me an afternoon to spot it even with help :-)
 *
 *       ROWS_child_lock locks this child and its containers, and its
 *       containers containers. It is generally always safe to call this.
 *       Under some circumstances, it is safe just to use ROWS_widget_lock
 *       on the child widget. (If you're only manipulating the child widget
 *       and not going to touch its container). Be cautious of mutex
 *       notifications under debugging (disable LinuxThreads extensions),
 *       as these could be the first signs of trouble. (You lock the child
 *       widget, do something that affects the container, ROWS_container
 *       calls ROWS_child_lock, which relocks the child, producing a mutex
 *       recursion notice)
 */
extern void ROWS_child_lock  (ROWS_child child);
extern void ROWS_child_unlock(ROWS_child child);

extern void ROWS_child_register_event(ROWS_child            child,
				      ROWS_child_event_num  event,
				      ROWS_child_func       func,
				      void                 *data);
extern void ROWS_child_event_send    (ROWS_child            child,
				      ROWS_child_event_num  event,
				      ROWS_child_event     *data);


extern inline ROWS_style ROWS_child_get_style(ROWS_child child);

extern void               ROWS_child_set_size  (ROWS_child  child,
						ROWS_point *size);
extern inline ROWS_point *ROWS_child_get_size  (ROWS_child  child);
extern void               ROWS_child_set_border(ROWS_child  child,
						ROWS_point *border);
extern ROWS_point        *ROWS_child_get_border(ROWS_child  child);
extern void               ROWS_child_set_origin(ROWS_child  child,
						ROWS_point *origin);
extern ROWS_point        *ROWS_child_get_origin(ROWS_child  child);

extern void          ROWS_child_register_drawable(ROWS_child    child,
						  ROWS_drawable drawable);
extern ROWS_drawable ROWS_child_get_drawable     (ROWS_child    child);

extern void ROWS_child_show          (ROWS_child child);
extern void ROWS_child_hide          (ROWS_child child);
extern void ROWS_child_realise       (ROWS_child child);
extern void ROWS_child_unrealise     (ROWS_child child);
extern void ROWS_child_set_focus     (ROWS_child child);
extern void ROWS_child_lose_focus    (ROWS_child child);
extern void ROWS_child_inactive      (ROWS_child child,
				      impulse_int inactive);
extern void ROWS_child_actual_widget (ROWS_child child);

extern void ROWS_child_set_activation(ROWS_child             child,
				      ROWS_style_active_type type);
extern void ROWS_child_set_state     (ROWS_child             child,
				      ROWS_child_state       state,
				      impulse_int            grouping);
extern void ROWS_child_update        (ROWS_child             child);

extern ROWS_child_state       ROWS_child_get_state     (ROWS_child       child);
extern ROWS_style_active_type ROWS_child_get_activation(ROWS_child       child);
extern ROWS_child             ROWS_child_get_selected  (ROWS_child_state state);

#endif
