/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the ROWS widget library
 * for Impulse.
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * ROWS_Xcomm.c
 * --------
 *
 * ROWS X-specific communications routines
 */

#include "impulse/rows/comms_int.h"

#include "impulse/util/memory.h"
#include "impulse/util/error.h"
#include "impulse/util/thread.h"
#include "impulse/util/event.h"
#include "impulse/rows/main_xwin.h"

#include <X11/Xlib.h>
#include <X11/extensions/sync.h>

static Window comm_window;
#if USE_THREADS
static impulse_mutex_static comm_lock = IMPULSE_MUTEX_CREATE_STATIC;
#endif


struct _ROWS_comms
{
  impulse_unistring name;
  Atom data;
  Atom control;
};


struct _ROWS_comms_control
{
  impulse_bool in_use;
  impulse_int pending;
};


typedef struct _ROWS_comms *ROWS_comms;
typedef struct _ROWS_comms_control *ROWS_comms_control;



static impulse_bool window_event(impulse_event *event,
				 void *data)
{
  XEvent *xevent = (XEvent *)event->data.system.data;

//  printf("Event %x %d\n", xevent->xany.window, xevent->type);
  switch (xevent->type)
  {
   case PropertyNotify:
    break;
    
   default:
    break;
  }
  impulse_free(xevent);
  
  return 1;
}

		    
void ROWS_comms_init(void)
{
#if USE_THREADS
  impulse_mutex_lock(&comm_lock);
#endif

  comm_window = XCreateWindow(appl->display,
			      appl->screen_window,
			      0, 0,
			      1, 1,
			      0,
			      0,
			      InputOnly,
			      CopyFromParent,
			      0,
			      NULL);

  XSelectInput(appl->display,
	       comm_window,
	       PropertyChangeMask);
  
  impulse_event_register(IMPULSE_EVENT_GUI,
			 comm_window,
			 window_event,
			 NULL);

#if USE_THREADS
  impulse_mutex_unlock(&comm_lock);
#endif
}


void ROWS_comms_cleanup(void)
{
#if USE_THREADS
  impulse_mutex_lock(&comm_lock);
#endif

  impulse_event_deregister(IMPULSE_EVENT_GUI,
			   comm_window,
			   NULL);
  
  XDestroyWindow(appl->display, comm_window);
  
#if USE_THREADS
  impulse_mutex_unlock(&comm_lock);
#endif
}


void *ROWS_comms_create(void)
{
  ROWS_comms comm = impulse_alloc(sizeof(struct _ROWS_comms));
  
  comm->name = NULL;
  comm->data = 0;
  comm->control = 0;
  
  return comm;
}


void ROWS_comms_destroy(void *data)
{
  ROWS_comms comm = (ROWS_comms)data;
  
  if (comm->name)
    impulse_unistring_destroy(comm->name);
  
  impulse_free(comm);
}


impulse_bool ROWS_comms_is_open(void *data)
{
  ROWS_comms comm = (ROWS_comms)data;
  
  return comm->control != 0;
}


impulse_int ROWS_comms_open(void *data,
			     impulse_unistring address)
{
  ROWS_comms comm = (ROWS_comms)data;
  impulse_int offset = impulse_unistring_find_chr(address, (impulse_unichar)':');
  impulse_unistring cname;
  impulse_unistring dname;
  struct _ROWS_comms_control control;
							 
  if (comm->control)
    ROWS_comms_close(data);
  
  if (offset < 0)
    return -1;

  cname = impulse_unistring_create_utf("IMPULSE_COMMS_CONTROL_", 8);
  dname = impulse_unistring_create_utf("IMPULSE_COMMS_DATA_", 8);
  
  comm->name = impulse_unistring_create_empty();
  impulse_unistring_substring(comm->name, address, offset+1, address->length);
  
  impulse_unistring_append(cname, comm->name);
  impulse_unistring_append(dname, comm->name);
  
  comm->control = ROWS_X_atom_id(cname);
  comm->data = ROWS_X_atom_id(dname);
  
  impulse_unistring_destroy(cname);
  impulse_unistring_destroy(dname);

  /* Fix me: byte swapping */
  control.pending = 0;
  control.in_use = 1;
  XChangeProperty(appl->display,
		  comm_window,
		  comm->control,
		  comm->control,
		  8,
		  PropModeReplace,
		  (char *)&control,
		  sizeof(control));
		  
  return 0;
}


impulse_int ROWS_comms_close(void *data)
{
  ROWS_comms comm = (ROWS_comms)data;
  
  if (!comm->control)
    return -1;
  
  impulse_unistring_destroy(comm->name);
  comm->control = 0;
  comm->data = 0;
  
  return 0;
}


impulse_int ROWS_comms_handle(void *data)
{
  ROWS_comms comm = (ROWS_comms)data;
  Atom hack;
  
  /* Hack to stop compile generating xxx variable not used warnings */
  hack = comm->data;
  
  return -1;
}


impulse_int ROWS_comms_pending(void *data)
{
  ROWS_comms comm = (ROWS_comms)data;
  ROWS_comms_control control = NULL;
  Atom actual_type = None;
  impulse_int actual_format = 0;
  impulse_int nitems = 0;
  impulse_int after = 0;
  impulse_int pending = 0;
  
  if (XGetWindowProperty(appl->display, 
			 comm_window,
			 comm->control,
			 (long)0, 
			 (long)sizeof(struct _ROWS_comms_control),
			 False,
			 comm->control,
			 &actual_type,
			 &actual_format,
			 (unsigned long *)&nitems,
			 (unsigned long *)&after,
			 (unsigned char **)&control) != Success)
  {
    return 0;
  }

  /* Fix me: byte swap */
  pending = control->pending;
  XFree(control);
  return pending;
}


impulse_int ROWS_comms_send(void *data,
			     impulse_message msg,
			     impulse_unistring address)
{
}


impulse_int ROWS_comms_receive(void *data,
				impulse_message *msg,
				impulse_unistring *address)
{
}


impulse_int ROWS_comms_address(void *data,
				impulse_unistring *address)
{
  ROWS_comms comm = (ROWS_comms)data;
  impulse_char buffer[1024];
  
  sprintf(buffer, "win:%x:", (int)comm_window);
  
  *address = impulse_unistring_create_utf(buffer, 8);
  impulse_unistring_append(*address, comm->name);

  return 0;
}



/* End of file */
