/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * container.c
 * -----------
 *
 * Class of widgets containing containers
 */

#include "../../../config.h"

#include "impulse/rows/container_int.h"
#include "impulse/rows/child_int.h"
#include "impulse/util/memory.h"

/*                            ======--------======                           */

static void ROWS_container_setup(ROWS_widget   *widget);
static int  container_ID   = -1;
static int  container_type = -1;

static void ROWS_container_destroy(void **private);

/* Contained events */
static int  _p_ROWS_container_realise(ROWS_child            widget,
				      ROWS_child_event_num  event_number,
				      ROWS_child_event     *event,
				      void                     *data);
static int  _p_ROWS_container_redraw (ROWS_child            widget,
				      ROWS_child_event_num  event_number,
				      ROWS_child_event     *event,
				      void                     *data);
static int  _p_ROWS_container_resize (ROWS_child            widget,
				      ROWS_child_event_num  event_number,
				      ROWS_child_event     *event,
				      void                     *data);
static int  _p_ROWS_container_click  (ROWS_child            widget,
				      ROWS_child_event_num  event_number,
				      ROWS_child_event     *event,
				      void                     *data);

/* Container events (confusing, ain't it?) */
static int  _p_ROWS_container_changed(ROWS_container            widget,
				      ROWS_container_event_num  event_number,
				      ROWS_container_event     *event,
				      void                     *data);
static int  _p_ROWS_container_size   (ROWS_container            widget,
				      ROWS_container_event_num  event_number,
				      ROWS_container_event     *event,
				      void                     *data);

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_container_create(ROWS_drawable drawable)
{
  ROWS_widget *wid;

  if (container_ID   == -1)
    container_ID   = ROWS_child_next_identifier();

  wid = ROWS_container_initialise(container_ID,
				  drawable);

  return *wid;
}

ROWS_widget *ROWS_container_initialise(int id,
				       ROWS_drawable drawable)
{
  ROWS_widget *widget;

  if (container_ID   == -1)
    container_ID   = ROWS_child_next_identifier();
  if (container_type == -1)
    container_type = ROWS_widget_next_type();

  /* Initialises the child widget underneath this one: */
  /* (Had me going for a minute there) */
  if (drawable != NULL)
    widget = ROWS_child_init_drawable(id, drawable);
  else
    widget = ROWS_child_initialise(id);
  
  ROWS_widget_set_type(widget, container_ID, container_type);
  ROWS_widget_set_destructor(widget, container_ID, ROWS_container_destroy);

  ROWS_container_setup(widget);

  return widget;
}

int ROWS_container_next_identifier(void)
{
  if (container_ID==-1)
    container_ID = ROWS_child_next_identifier();

  return container_ID + 1;
}

ROWS_container ROWS_container_cast(void **widget)
{
  if (container_ID   == -1)
    container_ID   = ROWS_child_next_identifier();
  if (container_type == -1)
    container_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, container_ID, container_type);
}

void ROWS_container_setup(ROWS_widget *widget)
{
  _p_ROWS_container_data **private;
  int                      x;
  ROWS_child           child;
  ROWS_container           container;

  private  = (_p_ROWS_container_data **) ROWS_widget_private(widget, container_ID);
  *private = impulse_alloc(sizeof(_p_ROWS_container_data));

  (*private)->inside = NULL;

  for (x=0; x<ROWS_CONTAINER_MAX_EVENT; x++)
    (*private)->functions[x] = NULL;

  child = ROWS_child_cast(*widget);
  container = ROWS_container_cast(*widget);

  ROWS_eventbox_register_widget(ROWS_eventbox_cast(*widget),
				container_ID,
				ROWS_CONTAINER_MAX_EVENT);
				
  ROWS_child_register_event(child,
				ROWS_CHILD_REALISE,
				_p_ROWS_container_realise,
				NULL);
  ROWS_child_register_event(child,
				ROWS_CHILD_REDRAW,
				_p_ROWS_container_redraw,
				NULL);
  ROWS_child_register_event(child,
			    ROWS_CHILD_RESIZE,
			    _p_ROWS_container_resize,
			    NULL);
  ROWS_child_register_event(child,
			    ROWS_CHILD_CLICK,
			    _p_ROWS_container_click,
			    NULL);

  ROWS_container_register_event(container,
				ROWS_CONTAINER_CHANGED,
				_p_ROWS_container_changed,
				NULL);
  ROWS_container_register_event(container,
				ROWS_CONTAINER_SIZE,
				_p_ROWS_container_size,
				NULL);
}

void ROWS_container_destroy(void **data)
{
  _p_ROWS_container_data **private;

  private  = (_p_ROWS_container_data **) data;

  impulse_free(*private);
}

/*                            ======--------======                           */

/* Class-specific functions */

/* Event functions */

void ROWS_container_register_event(ROWS_container           container,
				   ROWS_container_event_num event,
				   ROWS_container_func      func,
				   void                    *data)
{
  ROWS_eventbox_register_handler(ROWS_eventbox_cast(container),
				 container_ID,
				 (ROWS_eventbox_func) func,
				 event,
				 data);
 
  /* _p_ROWS_container_data **private;
  container_funclist      *next;

  ROWS_child_lock(container);
  ROWS_widget_check_type(container, container_type, "ROWS_container");

  private = (_p_ROWS_container_data **) container;
  next    = impulse_alloc(sizeof(container_funclist));

  next->next     = (*private)->functions[event];
  next->function = func;

  (*private)->functions[event] = next;
  ROWS_child_unlock(container); */
}

void ROWS_container_event_send(ROWS_container            container,
			       ROWS_container_event_num  event,
			       ROWS_container_event     *data)
{
  ROWS_widget_check_type(container, container_type, "ROWS_container");

  ROWS_child_lock(container);
  
  ROWS_eventbox_send(ROWS_eventbox_cast(container),
		     container_ID,
		     container,
		     event,
		     data);

  ROWS_child_unlock(container);
  
/*  _p_ROWS_container_data **private;
  container_funclist      *next;
  
  ROWS_child_lock(container);
  ROWS_widget_check_type(container, container_type, "ROWS_container");

  private = (_p_ROWS_container_data **) container;
  next    = (*private)->functions[event];

  while (next!=NULL)
    {
      if ((next->function)(container, event, data))
	{
	  ROWS_child_unlock(container);
	  return;
	}
      next = next->next;
    }
  ROWS_child_unlock(container); */
}

/* Informative functions about a child widget */

void ROWS_container_changed(ROWS_container container,
			    ROWS_child child)
{
  ROWS_container_event event;

  event.event.changed.changed = child;

  ROWS_container_event_send(container, ROWS_CONTAINER_CHANGED, &event);
}

void ROWS_container_size(ROWS_container container,
			 ROWS_child child)
{
  ROWS_container_event event;

  event.event.size.to_size = child;

  ROWS_container_event_send(container, ROWS_CONTAINER_SIZE, &event);
}

/* Addition/removal/status alteration functions */

void ROWS_container_add(ROWS_container container,
			ROWS_child child,
			int            padding,
			int            flags)
{
  ROWS_container_add_special(container, child, padding, flags, NULL);
}

/*                            ======--------======                           */

/* Responses to various events */

static int _p_ROWS_container_realise(ROWS_child            widget,
				     ROWS_child_event_num  event_number,
				     ROWS_child_event     *event,
				     void                     *data)
     /*
      * Upon realisation, give all the child widgets a drawable and then
      * realise them too.
      */
{
  ROWS_container             container;
  _p_ROWS_container_data   **private;
  ROWS_container_child  *next;
  ROWS_container_event       ev;

  /* ROWS_child_lock(widget); */
  container = ROWS_container_cast(widget);
  private   = (_p_ROWS_container_data **) container;

  next = (*private)->inside;

  while (next != NULL)
    {
      if (ROWS_child_get_drawable(next->item) == NULL)
	ROWS_child_register_drawable(next->item,
				     ROWS_child_get_drawable(widget));
      ROWS_child_realise(next->item);
      next = next->next;
    }

  ROWS_container_event_send(container,
			    ROWS_CONTAINER_SIZE,
			    &ev);
  /* ROWS_child_unlock(widget); */

  return 0; /* Allow control to continue */
}

static int _p_ROWS_container_redraw(ROWS_child            widget,
				    ROWS_child_event_num  event_number,
				    ROWS_child_event     *event,
				    void                     *data)
{
  _p_ROWS_container_data **private;
  _p_ROWS_child_data     **child;
  ROWS_container_child    *next;

  /* ROWS_child_lock(widget); */
  private   = (_p_ROWS_container_data **) ROWS_container_cast(widget);

  next = (*private)->inside;

  if (next!=NULL)
    {
      child = (_p_ROWS_child_data **) ROWS_child_cast(next->item);
      
      if ((*child)->shown)
	{
	  ROWS_child_event_send(next->item,
				    ROWS_CHILD_REDRAW,
				    event);

	  /* ROWS_child_unlock(widget); */

	  return 1; /* Redrawing must be complete now */
	}
    }

  /* ROWS_child_unlock(widget); */

  return 0; /* Allow control to continue */
}

static int _p_ROWS_container_click(ROWS_child            widget,
				   ROWS_child_event_num  event_number,
				   ROWS_child_event     *event,
				   void                 *data)
{
  _p_ROWS_container_data   **private;
  _p_ROWS_child_data   **child;
  ROWS_container_child  *next;

  ROWS_child_lock(widget);
  private   = (_p_ROWS_container_data **) ROWS_container_cast(widget);

  next = (*private)->inside;

  if (next != NULL)
    {
      ROWS_child_event ev;

      child = (_p_ROWS_child_data **) ROWS_child_cast(next->item);
      if (!(*child)->shown)
      {
	ROWS_child_unlock(widget);
	return 0;
      }
      ev = *event;
      ROWS_child_event_send((void **) child,
			    ROWS_CHILD_CLICK,
			    &ev);

      ROWS_child_unlock(widget);

      return 1;
    }

  ROWS_child_unlock(widget);

  return 0;
}

static int _p_ROWS_container_resize(ROWS_child            widget,
				    ROWS_child_event_num  event_number,
				    ROWS_child_event     *event,
				    void                 *data)
{
  _p_ROWS_container_data   **private;
  _p_ROWS_child_data   **child;
  ROWS_container_child  *next;

  ROWS_child_lock(widget);
  private   = (_p_ROWS_container_data **) ROWS_container_cast(widget);

  next = (*private)->inside;

  if (next!=NULL)
    {
      ROWS_child_event ev;
      
      child = (_p_ROWS_child_data **) ROWS_child_cast(next->item);

      ROWS_box_set_physical(ROWS_box_cast(next->item),
			    &event->event.resize.newsize);

      ev = *event;
      
      ROWS_child_event_send(ROWS_child_cast(next->item),
			    ROWS_CHILD_RESIZE,
			    &ev);

      ROWS_child_unlock(widget);

      return 1; /* We've resized the widget */
    }

  ROWS_child_unlock(widget);

  return 0;
}
static int _p_ROWS_container_changed(ROWS_container            widget,
				     ROWS_container_event_num  event_number,
				     ROWS_container_event     *event,
				     void                     *data)
{
  _p_ROWS_container_data   **private;
  _p_ROWS_child_data   **child;
  ROWS_container_child  *next;
  ROWS_child_event       ev;
  ROWS_rect              rect;

  ROWS_child_lock(widget);

  private   = (_p_ROWS_container_data **) widget;

  next = (*private)->inside;

  child = (_p_ROWS_child_data **) next->item;

  if (child!=NULL)
    {
      if (ROWS_drawable_get_valid((*child)->draw))
	{
	  /* Redraw the child widget */
	  ROWS_box_get_boundaries(ROWS_box_cast(next->item),
				  &rect);
      
	  ev.event.redraw.clip    = rect;
	  ev.event.redraw.invalid = rect;
      
	  ROWS_child_event_send(next->item,
				ROWS_CHILD_REDRAW,
				&ev);
	}
    }

  ROWS_child_unlock(widget);

  return 1; /* The buck stops here */
}

static int _p_ROWS_container_size(ROWS_container            widget,
				  ROWS_container_event_num  event_number,
				  ROWS_container_event     *event,
				  void                     *data)
{
  _p_ROWS_container_data   **private;
  ROWS_container_child      *next;
  ROWS_point                *size;
  ROWS_point                 bounding;
  ROWS_box                   box;

  ROWS_child_lock(widget);

  private   = (_p_ROWS_container_data **) ROWS_container_cast(widget);

  next = (*private)->inside;

  if (next!=NULL)
    {
      ROWS_child_realise        (next->item);
      size = ROWS_child_get_size(next->item);
      
      box      =  ROWS_box_cast(widget);
      bounding = *ROWS_box_get_virtual(box);
      
      bounding.x = 72000;
      bounding.y = 72000;
      
      if (size->x>0 && size->y>0)
	{
	  bounding.x = size->x;
	  bounding.y = size->y;
	}
      
      ROWS_box_set_boundaries(box, &bounding);
      ROWS_box_set_boundaries(ROWS_box_cast(next->item), &bounding);

      next->ideal_size.x = bounding.x;
      next->ideal_size.y = bounding.y;

      next = next->next;
    }
  else
    {
      box      =  ROWS_box_cast(widget);
      bounding = *ROWS_box_get_virtual(box);
      
      bounding.x = 72000;
      bounding.y = 72000;
     
      ROWS_box_set_boundaries(box, &bounding);
    }

  ROWS_child_unlock(widget);
      
  return 1; /* Control terminates */  
}

/*                            ======--------======                           */

void ROWS_container_add_special(ROWS_container      container,
				ROWS_child          child,
				int                 padding,
				int                 flags,
				void               *data)
{
  _p_ROWS_container_data   **private;
  ROWS_container_child      *next;
  _p_ROWS_child_data       **child_data;

  /* ROWS_widget_lock(container); */ /* No need to lock the container - child_lock does this for us */
  ROWS_child_lock(container);
  ROWS_widget_lock(child);
  ROWS_widget_check_type(container, container_type, "ROWS_container");

  private    = (_p_ROWS_container_data **) container;
  child_data = (_p_ROWS_child_data **) ROWS_child_cast(child);

  next = impulse_alloc(sizeof(ROWS_container_child));

  next->next    = (*private)->inside;
  next->item    = ROWS_child_cast(child);
  next->padding = padding;
  next->flags   = flags;
  next->data    = data;

  /* Generic ideal size */
  next->ideal_size.x = 72000;
  next->ideal_size.y = 72000;

  (*private)->inside = next;

  ROWS_child_register_drawable(child,
			       ROWS_child_get_drawable(ROWS_child_cast(container)));

  (*child_data)->container = ROWS_container_cast(container);

  ROWS_child_container(child, container);

  ROWS_widget_unlock(child);
  ROWS_child_unlock(container);
}

void ROWS_container_iterator(ROWS_container       container,
			     ROWS_container_iter  function,
			     void                *moredata)
{
  _p_ROWS_container_data   **private;
  ROWS_container_child      *next;

  ROWS_widget_check_type(container, container_type, "ROWS_container");
  ROWS_child_lock(container);

  private   = (_p_ROWS_container_data **) container;

  next = (*private)->inside;

  while (next!=NULL)
    {
      (*function)(container,
		  next->item,
		  next->data,
		  moredata);

      next = next->next;
    }
  
  ROWS_child_unlock(container);
}

void *ROWS_container_lookup(ROWS_container       container,
			    ROWS_child           child)
{
   _p_ROWS_container_data   **private;
  ROWS_container_child      *next;
  void                      *data;

  ROWS_widget_check_type(container, container_type, "ROWS_container");
  /* ROWS_widget_lock(container); */
  ROWS_child_lock(child);

  private = (_p_ROWS_container_data **) container;

  next = (*private)->inside;
  data = NULL;

  while (next!=NULL)
    {
      if (next->item == child)
	data = next->data;

      next = next->next;
    }

  ROWS_child_unlock(child);

  return data;
}
