/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * drawable.c
 * ----------
 *
 * A box that can be drawn on
 */

#include "../../../config.h"

#include "impulse/rows/drawable_int.h"
#include "impulse/util/memory.h"
#include "impulse/util/error.h"

/*                            ======--------======                           */

static void ROWS_drawable_setup(ROWS_widget *widget);
static int  drawable_ID   = -1;
static int  drawable_type = -1;

static void ROWS_drawable_destroy(void **private);

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_drawable_create(void)
{
  ROWS_widget *wid;

  if (drawable_ID == -1)
    drawable_ID   = ROWS_box_next_identifier();

  wid = ROWS_drawable_initialise(drawable_ID);

  return *wid;
}

ROWS_widget *ROWS_drawable_initialise(int id)
{
  ROWS_widget *widget;

  if (drawable_ID   == -1)
    drawable_ID   = ROWS_box_next_identifier();
  if (drawable_type == -1)
    drawable_type = ROWS_widget_next_type();

  widget = ROWS_box_initialise(id);
  ROWS_widget_set_type(widget, drawable_ID, drawable_type);
  ROWS_widget_set_destructor(widget, drawable_ID, ROWS_drawable_destroy);

  ROWS_drawable_setup(widget);

  return widget;
}

int ROWS_drawable_next_identifier(void)
{
  if (drawable_ID==-1)
    drawable_ID = ROWS_box_next_identifier();

  return drawable_ID + 1;
}

ROWS_drawable ROWS_drawable_cast(void **widget)
{
  if (drawable_ID   == -1)
    drawable_ID   = ROWS_box_next_identifier();
  if (drawable_type == -1)
    drawable_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, drawable_ID, drawable_type);
}

static void ROWS_drawable_invalid_function(ROWS_drawable_fparam *param)
{
  impulse_unistring desc = impulse_unistring_create_utf("Unsupported drawable function", 8);
  impulse_unistring msg = impulse_unistring_create_utf("Are you using an old drawable library with a newer version of ROWS?", 8);
  impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, msg);
}

void ROWS_drawable_setup(ROWS_widget *widget)
{
  _p_ROWS_drawable_data **private;
  int                     x;

  private  = (_p_ROWS_drawable_data **) ROWS_widget_private(widget, drawable_ID);
  *private = impulse_alloc(sizeof(_p_ROWS_drawable_data));

  (*private)->last_brush  = NULL;
  (*private)->clip_serial = 0;
  (*private)->use_region  = 0;

  for (x=0;x<ROWS_DRAWABLE_MAX_FUNC;x++)
    (*private)->abstraction_layer.function[x] = ROWS_drawable_invalid_function;

  (*private)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_CREATE]  = NULL;
  (*private)->abstraction_layer.function[ROWS_DRAWABLE_BRUSH_DESTROY] = NULL;

  (*private)->style = NULL;
}

void ROWS_drawable_postinitialise(ROWS_widget drawable,
				  int         has_styles)
{
  _p_ROWS_drawable_data **private;

  private  = (_p_ROWS_drawable_data **) ROWS_drawable_cast(drawable);

  if (has_styles)
    (*private)->style = ROWS_style_create((ROWS_drawable) private);
}

void ROWS_drawable_destroy(void **private)
{
  _p_ROWS_drawable_data **data;

  data = (_p_ROWS_drawable_data **) private;

  if ((*data)->style != NULL)
    ROWS_style_dereference((*data)->style);

  impulse_free(*private);
}

/*                            ======--------======                           */

/* Class-specific functions */

void ROWS_drawable_set_valid(ROWS_drawable drawable,
			     int           valid)
{
  _p_ROWS_drawable_data **private;

  ROWS_widget_lock(drawable);
  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private = (_p_ROWS_drawable_data **) drawable;

  (*private)->valid = (valid!=0);
  ROWS_widget_unlock(drawable);
}

int ROWS_drawable_get_valid(ROWS_drawable drawable)
{
  _p_ROWS_drawable_data **private;
  
  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private = (_p_ROWS_drawable_data **) drawable;

  return (*private)->valid;
}

/*                            ======--------======                           */

/* Style (theme) functions */

ROWS_style ROWS_drawable_get_style(ROWS_drawable drawable)
{
  _p_ROWS_drawable_data **private;

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;
  return (*private)->style;
}

/*                            ======--------======                           */

/* Abstraction functions */

extern ROWS_drawable_brush_calls *ROWS_drawable_get_abstraction_layer(ROWS_drawable drawable)
{
  _p_ROWS_drawable_data **private;

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;
  return &(*private)->abstraction_layer;
}

/*                            ======--------======                           */

/* Clipping rectangle functions */

int ROWS_drawable_get_clip_serial(ROWS_drawable drawable)
{
  _p_ROWS_drawable_data **private;

#ifdef ROWS_DEBUG
  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");
#endif

  private  = (_p_ROWS_drawable_data **) drawable;
  return (*private)->clip_serial;
}

void ROWS_drawable_set_clip(ROWS_drawable  drawable,
			    ROWS_rect     *clip)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam    param;

  ROWS_widget_lock(drawable);

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;

  (*private)->clip_serial++;

  param.drawable.drawable       =  drawable;
  param.drawable.data.clip.rect = *clip;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_SET_CLIP])
    (&param);

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_clip_region(ROWS_drawable drawable,
			       int           use_region)
{
  _p_ROWS_drawable_data **private;

  ROWS_widget_lock(drawable);

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;

  (*private)->clip_serial++;
  (*private)->use_region = use_region;

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_clip_push(ROWS_drawable drawable)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam    param;

  ROWS_widget_lock(drawable);

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;

  param.drawable.drawable   = drawable;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_CLIP_PUSH])
    (&param);

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_clip_pop(ROWS_drawable drawable)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam    param;

  ROWS_widget_lock(drawable);

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;

  param.drawable.drawable   = drawable;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_CLIP_POP])
    (&param);
  (*private)->clip_serial++;

  ROWS_drawable_clip_region_set(drawable);

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_clip_region_clear(ROWS_drawable drawable)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam    param;

  ROWS_widget_lock(drawable);

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;

  param.drawable.drawable   = drawable;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_CLIP_REGION_CLEAR])
    (&param);

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_clip_region_add(ROWS_drawable  drawable,
				   ROWS_rect     *clip)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam    param;

  ROWS_widget_lock(drawable);

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;

  (*private)->clip_serial++;

  param.drawable.drawable       = drawable;
  param.drawable.data.clip.rect = *clip;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_CLIP_REGION_ADD])
    (&param);

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_clip_region_set(ROWS_drawable drawable)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam    param;

  ROWS_widget_lock(drawable);

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;

  param.drawable.drawable       = drawable;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_CLIP_REGION_SET])
    (&param);

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_clip_region_bounds(ROWS_drawable  drawable,
				      ROWS_rect     *rect)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam    param;

  ROWS_widget_lock(drawable);

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;
  
  param.drawable.drawable = drawable;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_CLIP_REGION_BOUNDS])
    (&param);

  *rect = param.drawable.data.clip.rect;

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_clip_region_move(ROWS_drawable  drawable,
				    ROWS_point    *offset)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam    param;

  ROWS_widget_lock(drawable);

  ROWS_widget_check_type(drawable, drawable_type, "ROWS_drawable");

  private  = (_p_ROWS_drawable_data **) drawable;

  (*private)->clip_serial++;

  param.drawable.drawable             =   drawable;
  param.drawable.data.copy.offset     =  *offset;
  param.drawable.data.copy.use_region = (*private)->use_region;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_CLIP_REGION_MOVE])
    (&param);

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_clip_region_copy(ROWS_drawable  to,
				    ROWS_drawable  from,
				    ROWS_point    *offset)
{
  _p_ROWS_drawable_data **private, **private_from;
  ROWS_drawable_fparam    param;

  ROWS_widget_lock(to);
  ROWS_widget_lock(from);

  ROWS_widget_check_type(to,   drawable_type, "ROWS_drawable");
  ROWS_widget_check_type(from, drawable_type, "ROWS_drawable");

  private      = (_p_ROWS_drawable_data **) to;
  private_from = (_p_ROWS_drawable_data **) from;

  (*private)->clip_serial++;

  (*private)->use_region = (*private_from)->use_region;

  param.drawable.drawable             =   to;
  param.drawable.data.copy.from       =   from;
  param.drawable.data.copy.use_region = (*private)->use_region;
  param.drawable.data.copy.offset     =  *offset;
  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_CLIP_REGION_COPY])
    (&param);

  ROWS_widget_unlock(to);
  ROWS_widget_unlock(from);
}

/*                            ======--------======                           */

/* Drawing functions */

void ROWS_drawable_clear_area(ROWS_drawable drawable,
			      int x, int y,
			      int w, int h)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam    param;

  ROWS_widget_lock(drawable);

  private = (_p_ROWS_drawable_data **) drawable;

  param.drawable.drawable = drawable;
  param.drawable.data.clear_area.x = x;
  param.drawable.data.clear_area.y = y;
  param.drawable.data.clear_area.w = w;
  param.drawable.data.clear_area.h = h;

  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_CLEAR_AREA])(&param);

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_sync(ROWS_drawable drawable)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam param;

  ROWS_widget_lock(drawable);

  private = (_p_ROWS_drawable_data **) drawable;

  param.drawable.drawable = drawable;

  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_SYNC])(&param);

  ROWS_widget_unlock(drawable);
}

void ROWS_drawable_draw_bitmap(ROWS_drawable drawable,
			       ROWS_bitmap   bitmap,
			       microns       x,
			       microns       y)
{
  _p_ROWS_drawable_data **private;
  ROWS_drawable_fparam param;

  ROWS_widget_lock(drawable);

  private = (_p_ROWS_drawable_data **) drawable;

  param.drawable.drawable = drawable;
  param.drawable.data.draw_bitmap.bitmap = bitmap;
  param.drawable.data.draw_bitmap.x      = x;
  param.drawable.data.draw_bitmap.y      = y;

  ((*private)->abstraction_layer.function[ROWS_DRAWABLE_DRAW_BITMAP])(&param);

  ROWS_widget_unlock(drawable);
}
