/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Danie Berrange
 * are Copyright (C) 1999 Daniel Berrange
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * error.c
 * -------
 *
 * Impulse program error handling
 */

#include "../../../config.h"

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include <syslog.h>
#include <signal.h>

#include "impulse/util/error.h"
#include "impulse/util/memory.h"
#include "impulse/util/thread.h"

struct __p_error_info
{
  char *msg;
  impulse_error_action base;
  impulse_error_action action;
  impulse_error_handler handler;
};

typedef struct __p_error_info _p_error_info;

static _p_error_info _p_info[] =
{
  (_p_error_info){
    "Emergency",
    IMPULSE_ERROR_EXIT,
    IMPULSE_ERROR_PRINT,
    NULL
  },
  (_p_error_info){
    "Critical",
    IMPULSE_ERROR_EXIT,
    IMPULSE_ERROR_PRINT,
    NULL
  },
  (_p_error_info){
    "Warning",
    IMPULSE_ERROR_NONE,
    IMPULSE_ERROR_PRINT,
    NULL
  },
  (_p_error_info){
    "Infomation",
    IMPULSE_ERROR_NONE,
    IMPULSE_ERROR_NONE,
    NULL
  }
};


extern impulse_unistring prog_path;


void _p_impulse_error_print(impulse_error_level level,
			    impulse_unistring desc,
			    impulse_unistring msg)
{
  char *cdesc = (desc ? impulse_unistring_to_utf(desc, 8, 0) : NULL);
  char *cmsg = (msg ? impulse_unistring_to_utf(msg, 8, 0) : NULL);
  
  fprintf(stderr, "*****\n");
  fprintf(stderr, "***** Impulse %s\n", _p_info[level].msg);
  if (desc)
    fprintf(stderr, "***** %s\n", cdesc);
  if (msg)
    fprintf(stderr, "***** %s\n", cmsg);
  fprintf(stderr, "*****\n");
  if (msg)
    impulse_free(cmsg);
  if (desc)
    impulse_free(cdesc);
}


static int _p_syslevel[] =
{
  LOG_EMERG,
  LOG_CRIT,
  LOG_WARNING,
  LOG_INFO
};


void _p_impulse_error_log(impulse_error_level level,
			  impulse_unistring desc,
			  impulse_unistring msg)
{
  char *cdesc = (desc ? impulse_unistring_to_utf(desc, 8, 0) : NULL);
  char *cmsg = (msg ? impulse_unistring_to_utf(msg, 8, 0) : NULL);
  
  openlog("Impulse", LOG_PID, LOG_USER);
  
  syslog(_p_syslevel[level], _p_info[level].msg);
  if (cdesc)
    syslog(_p_syslevel[level], cdesc);
  if (cmsg)
    syslog(_p_syslevel[level], cmsg);

  if (cmsg)
    impulse_free(cmsg);
  if (cdesc)
    impulse_free(cdesc);
}

#if ! USE_THREADS
static int child_died = 0;

static void child_handler(int sig)
{
  printf("Got sigchild\n");
  child_died = 1;
}
#endif

void _p_impulse_error_trace(void)
{
  /*
   * Fix Me:
   * This code works fine until threads are involved, at which
   * point all hell breaks loose. Don't ask me why, somebody
   * please fix it
   * 
   */
#if USE_THREADS
  printf("Stack traces are not working yet on multi-threaded systems\n");
#else
  int child;
  int new_stdin[2];
  int new_stdout[2];
  char pid[50];
  char *prog = impulse_unistring_to_utf(prog_path, 8, 0);
  char *args[5];
  struct sigaction action;
  struct sigaction oldaction;
  
  snprintf(pid, 49, "%d", getpid());
  printf("Foo %s\n", pid);
  args[0] = "gdb";
  args[1] = "-n";
  args[2] = prog;
  args[3] = pid;
  args[4] = NULL;

  if (pipe(new_stdin) < 0)
    return;
  
  if (pipe(new_stdout) < 0)
  {
    close(new_stdin[0]);
    close(new_stdin[1]);
  }

  sigfillset(&action.sa_mask);
  action.sa_handler   = child_handler;
  action.sa_flags     = 0;
  sigaction(SIGCHLD, &action, &oldaction);
  child_died = 0;

  child = fork();
  if (!child)
  {
    dup2(new_stdin[0], STDIN_FILENO);
    dup2(new_stdout[1], STDOUT_FILENO);
    dup2(new_stdout[1], STDERR_FILENO);
    
    execvp(args[0], args);
    /* Should never reach here */
    _exit(1);
  }
  else if (child < 0)
    return; /* Do something better*/

  write(new_stdin[1], "backtrace\n", 10);
  write(new_stdin[1], "quit\n", 5);
  write(new_stdin[1], "y\n", 2);

  while (!child_died)
  {
    char c;
    if (read(new_stdout[0], &c, 1) != 1)
      break;
    write(STDOUT_FILENO, &c, 1);
  }

  waitpid(child, NULL, 0);

  close(new_stdout[0]);
  close(new_stdout[1]);
  close(new_stdin[0]);
  close(new_stdin[1]);
  impulse_free(prog);
#endif
}


void _p_impulse_error_debug(void)
{
  /*
   * Fix Me:
   * See comment in stack trace 
   */
#if USE_THREADS
  printf("Debugging not working yet on multithreaded systems\n");
#else
  int child;
  char pid[50];
  char *prog = impulse_unistring_to_utf(prog_path, 8, 0);
  char *args[5];
  
  snprintf(pid, 49, "%d", getpid());
  
  args[0] = "gdb";
  args[1] = "-q";
  args[2] = prog;
  args[3] = pid;
  args[4] = NULL;
  
  child = fork();
  if (!child)
  {
    execvp(args[0], args);
    /* Should never reach here */
    _exit(1);
  }
  else if (child < 0)
    return; /* Do something better*/

  waitpid(child, NULL, 0);
  impulse_free(prog);
#endif
}



/* 
 * Fix me: internationalise error messages 
 */

#if USE_THREADS
static impulse_bool impulse_dead = 0;
static impulse_mutex_static error_mutex = IMPULSE_MUTEX_CREATE_STATIC;
#endif

void impulse_error_raise(impulse_error_level level,
			impulse_unistring desc,
			impulse_unistring msg)
{
  impulse_error_action action;
  
  if (level < 0 || level > IMPULSE_ERROR_MAX)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Unknown error level requested", 8);
    impulse_error_raise(IMPULSE_ERROR_WARNING, desc, NULL);
    impulse_unistring_destroy(desc);
    return;
  }
#if USE_THREADS
  impulse_mutex_lock(&error_mutex);
  if (impulse_dead)
  {
    /* 
     * If another thread has already killed the process, then
     * we might as well get the hell out as well
     */
    impulse_mutex_unlock(&error_mutex);
    _exit(1);
  }
#endif
  
  action = _p_info[level].action | _p_info[level].base;
  if (action & IMPULSE_ERROR_HANDLER)
    (*_p_info[level].handler)(level, desc, msg);
  if (action & IMPULSE_ERROR_PRINT)
    _p_impulse_error_print(level, desc, msg);
  if (action & IMPULSE_ERROR_LOG)
    _p_impulse_error_log(level, desc, msg);
  if (action & IMPULSE_ERROR_TRACE)
    _p_impulse_error_trace();
  if (action & IMPULSE_ERROR_DEBUG)
    _p_impulse_error_debug();
  if (action & IMPULSE_ERROR_EXIT)
  {
    /* Do some application cleanup */
#if USE_THREADS
    impulse_dead = 1;
    impulse_mutex_unlock(&error_mutex);
#endif
    exit(1);
  }
#if USE_THREADS
  impulse_mutex_unlock(&error_mutex);
#endif
}



void impulse_error_handle(impulse_error_level level,
			  impulse_error_action action,
			  impulse_error_handler handler)
{
  if (level < 0 || level > IMPULSE_ERROR_MAX)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Unknown error level requested", 8);
    impulse_error_raise(IMPULSE_ERROR_WARNING, desc, NULL);
    impulse_unistring_destroy(desc);
    return;
  }

#if USE_THREADS
  impulse_mutex_lock(&error_mutex);
#endif
  _p_info[level].action = action;
  _p_info[level].handler = handler;
#if USE_THREADS
  impulse_mutex_unlock(&error_mutex);
#endif
}


void impulse_error_syscall(impulse_error_level level,
			   impulse_unistring desc)
{
/*  char buffer[1024];
  impulse_unistring msg;
  
  strerror_r(errno, buffer, 1023);
  msg = impulse_unistring_create_utf(buffer, 8);

  impulse_error_raise(level, desc, msg);  
  impulse_unistring_destroy(msg);*/
  
  impulse_unistring msg = impulse_unistring_create_utf(strerror(errno), 8);
  impulse_error_raise(level, desc, msg);
  impulse_unistring_destroy(msg);
}


/* End of file */
