/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the ROWS widget library
 * for Impulse.
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 */

/*
 * event.h
 * -------
 *
 * Generalised event handling for specific architectures
 */

#ifndef __impulse_util_event_h
#define __impulse_util_event_h

#include <impulse/util/setup.h>
#include <sys/time.h>

/*                            ======--------======                           */

/* Structure of a impulse_event */

typedef struct _impulse_event {
  enum {
    IMPULSE_EVENT_SYSTEM,         /* System-specific event type */
    IMPULSE_EVENT_FILEDESCRIPTOR, /* File descriptor event      */
    IMPULSE_EVENT_ALARM,          /* Alarm event                */
    IMPULSE_EVENT_GUI,            /* An event range especially for GUIs */
    
    IMPULSE_EVENT_USER,           /* User defined event         */

    IMPULSE_EVENT_USER_MAX = IMPULSE_EVENT_USER + 16,

    IMPULSE_EVENT_MAX
  } type;

  impulse_int id;
  
  union {
    struct {
      impulse_int object;
    } any;     /* Generic event type */
    
    struct {
      impulse_int object;
      impulse_int type;
      void  *data;
    } system;      /* System-specific event data */

    struct {
      impulse_int fd;
      enum {
	IMPULSE_EVENT_FD_READ = 0x1,
	IMPULSE_EVENT_FD_WRITE = 0x2,
	IMPULSE_EVENT_FD_EXCEPTION = 0x4
      } type;
    } fd;          /* Filedescriptor event */

    struct {
      impulse_int timer;
      struct timeval set_for;
      struct timeval went_off;
    } alarm;       /* Alarm event */

    struct 
    {
      impulse_int object;
      impulse_int type;
      void  *data;
    } gui;      /* Gui stuff */
    
    struct {
      impulse_int object;
      impulse_int type;
      void *data;
    } user;     /* User defined */
  } data;
} impulse_event;

typedef impulse_bool (*impulse_event_handler)(impulse_event *event,
					      void *data);
typedef impulse_int (*impulse_event_hook)(void);

typedef impulse_int (*impulse_event_predicate)(impulse_event *event,
					       void *data);

/*                            ======--------======                           */

/* Low-level event handling stuff */

/* Event handling functions */
extern impulse_int impulse_event_fd(impulse_int fd, 
					impulse_int state);
extern impulse_int impulse_event_alarm(impulse_int timer,
				       impulse_int secs,
				       impulse_int usecs);
extern impulse_int impulse_event_alarm_offset(impulse_int timer,
					      impulse_int secs,
					      impulse_int usecs);

/* Triggering events */
/* Don't you dare call this unless you own the event mutex
 * ie from a preselect handler.
 */
extern void impulse_event_occured(impulse_event *event);


/*  Collecting events */
extern impulse_event *impulse_event_collect(void);

/* Routine to be run just before & after invoking select... */
/*
 * Note to self: in order to generate events, these must call
 * event_occured, *not* event_trigger. Everything else must call
 * event_trigger, *not* event_occured...
 */
extern void impulse_event_set_preselect_hook(impulse_event_hook routine);
extern void impulse_event_set_postselect_hook(impulse_event_hook routine);



/*                            ======--------======                           */
/*
 * High level event handling stuff
 */

/* Generic event handlers */
extern void impulse_event_register(impulse_int event_type,
				   impulse_int object,
				   impulse_event_handler func,
				   void *data);
extern void impulse_event_deregister(impulse_int event_type,
				     impulse_int obbject,
				     void **data);


/* File descriptors handlers */
extern void impulse_event_register_fd(impulse_int fd,
				      impulse_int state,
				      impulse_event_handler func,
				      void *data);
extern void impulse_event_deregister_fd(impulse_int fd);

/* Alarm handlers */
extern void impulse_event_register_alarm(impulse_int timer,
					 impulse_int secs,
					 impulse_int usecs,
					 impulse_event_handler func,
					 void *data);
extern void impulse_event_register_alarm_offset(impulse_int timer,
						impulse_int secs,
						impulse_int usecs,
						impulse_event_handler func,
						void *data);
extern void impulse_event_deregister_alarm(impulse_int id);

extern impulse_int impulse_event_timer(void);

/* Triggering events */
extern void impulse_event_trigger(impulse_event *event);


/*  Collecting events */
extern impulse_int impulse_event_pending(impulse_event_predicate pred,
					 void *data);
extern impulse_int impulse_event_check_next(impulse_event *result,
					    impulse_event_predicate pred,
					    void *data);
extern void impulse_event_get_next(impulse_event *result);
extern void impulse_event_get_next_predicate(impulse_event *result,
					     impulse_event_predicate pred,
					     void *data,
					     impulse_int priority);
extern void impulse_event_handle(void);

#endif
