/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 */

/*
 * main.c
 * ------
 *
 * Impulse program startup and termination functions
 */

#include "../../../config.h"

#include "impulse/util/main.h"
#include "impulse/util/memory.h"
#include "impulse/util/error.h"
#include "impulse/util/event.h"
#include "impulse/util/hash.h"
#include "impulse/util/thread.h"
#include "impulse/util/sparse.h"
#include "impulse/util/list.h"

#include <signal.h>
#include <stdlib.h>
#include <stdio.h>


struct _impulse_command_details
{
  impulse_unistring lopt;
  impulse_unichar sopt;
  impulse_unistring arg;
  enum _impulse_command_type type;
  impulse_unistring help;
  
  impulse_uint count;
  union _impulse_command_value *values;
};

typedef struct _impulse_command_details *impulse_command_details;

/*
 * Fix Me:
 * This is just a temporary hack to allow stack traces to work
 */

impulse_unistring prog_path = NULL;
impulse_unistring prog_name = NULL;

static impulse_unistring *uniargv = NULL;
static impulse_int uniargc = 0;
static impulse_int lastarg = 0;
static impulse_hash command_lopts = NULL;
static impulse_sparse command_sopts = NULL;
#if USE_THREADS
static impulse_mutex_static command_opts_lock = IMPULSE_MUTEX_CREATE_STATIC;
#endif

static impulse_list command_hooks = NULL;
#if USE_THREADS
static impulse_mutex_static command_hooks_lock = IMPULSE_MUTEX_CREATE_STATIC;
#endif


/*                            ======--------======                           */


#if HAVE_SIGINFO
static void _p_sighandler(int sig, siginfo_t *info, void *data)
#else
static void _p_sighandler(int sig)
#endif
{
  impulse_unistring desc = NULL;
  impulse_unistring msg = NULL;
  impulse_error_level level = IMPULSE_ERROR_EMERGENCY;

  switch (sig)
  {
   case SIGHUP:
    level = IMPULSE_ERROR_CRITICAL;
    desc = impulse_unistring_create_utf("Hang up from terminal", 8);
    break;
    
   case SIGINT:
    level = IMPULSE_ERROR_CRITICAL;
    desc = impulse_unistring_create_utf("Interrupted from terminal", 8);
    break;
    
   case SIGQUIT:
    level = IMPULSE_ERROR_CRITICAL;
    desc = impulse_unistring_create_utf("Quit from terminal", 8);
    break;
    
   case SIGILL:
    level = IMPULSE_ERROR_EMERGENCY;
    desc = impulse_unistring_create_utf("Illegal instruction", 8);

#if HAVE_SIGINFO
    switch (info->si_code)
    {
     case ILL_ILLOPC:
      msg = impulse_unistring_create_utf("Illegal opcode", 8);
      break;
     case ILL_ILLOPN:
      msg = impulse_unistring_create_utf("Illegal operand", 8);
      break;
     case ILL_ILLTRP:
      msg = impulse_unistring_create_utf("Illegal trap", 8);
      break;
     case ILL_ILLADR:
      msg = impulse_unistring_create_utf("Illegal addressing mode", 8);
      break;
     case ILL_PRVOPC:
      msg = impulse_unistring_create_utf("Priviledge operation in non-priviledged mode", 8);
      break;
     case ILL_PRVREG:
      msg = impulse_unistring_create_utf("Priviledged register in non-priviledged mode", 8);
      break;
     case ILL_COPROC:
      msg = impulse_unistring_create_utf("Coprocessor error", 8);
      break;
     case ILL_BADSTK:
      msg = impulse_unistring_create_utf("Internal stack error", 8);
      break;
    }
#endif
    break;
    
   case SIGABRT:
    level = IMPULSE_ERROR_EMERGENCY;
    desc = impulse_unistring_create_utf("Abnormal program termination", 8);
    break;
    
   case SIGFPE:
    level = IMPULSE_ERROR_CRITICAL;
    desc = impulse_unistring_create_utf("Floating point exception", 8);
    
#if HAVE_SIGINFO
    switch (info->si_code)
    {
     case FPE_INTDIV:
      msg = impulse_unistring_create_utf("Integer divide by zero", 8);
      break;
     case FPE_INTOVF:
      msg = impulse_unistring_create_utf("Integer overflow", 8);
      break;
     case FPE_FLTDIV:
      msg = impulse_unistring_create_utf("Division by zero", 8);
      break;
     case FPE_FLTOVF:
      msg = impulse_unistring_create_utf("Overflow", 8);
      break;
     case FPE_FLTUND:
      msg = impulse_unistring_create_utf("Underflow", 8);
      break;
     case FPE_FLTRES:
      msg = impulse_unistring_create_utf("Inexact result", 8);
      break;
     case FPE_FLTINV:
      msg = impulse_unistring_create_utf("Invalid operation", 8);
      break;
     case FPE_FLTSUB:
      msg = impulse_unistring_create_utf("Subscript out of range", 8);
      break;
    }
#endif
    break;
    
   case SIGSEGV:
    level = IMPULSE_ERROR_EMERGENCY;
    desc = impulse_unistring_create_utf("Segmentation fault", 8);
    
#if HAVE_SIGINFO
    switch(info->si_code)
    {
     case SEGV_MAPERR:
      msg = impulse_unistring_create_utf("Address not mapped", 8);
      break;
     case SEGV_ACCERR:
      msg = impulse_unistring_create_utf("Invalid permissions", 8);
      break;
    }
#endif
    break;
    
   case SIGPIPE:
    level = IMPULSE_ERROR_WARNING;
    desc = impulse_unistring_create_utf("Broken pipe", 8);
    break;
    
   case SIGALRM:
    level = IMPULSE_ERROR_WARNING;
    desc = impulse_unistring_create_utf("Alarm clock", 8);
    break;
    
   case SIGTERM:
    level = IMPULSE_ERROR_CRITICAL;
    desc = impulse_unistring_create_utf("Explicitly killed", 8);
    break;

   case SIGBUS:
    level = IMPULSE_ERROR_EMERGENCY;
    desc = impulse_unistring_create_utf("Bus error", 8);
    
#if HAVE_SIGINFO
    switch(info->si_code)
    {
     case BUS_ADRALN:
      msg = impulse_unistring_create_utf("Invalid alignment", 8);
      break;
     case BUS_ADRERR:
      msg = impulse_unistring_create_utf("Non-existant physical address", 8);
      break;
     case BUS_OBJERR:
      msg = impulse_unistring_create_utf("Object-specific hardware error", 8);
      break;
    }
#endif
    break;
    
   default:
    level = IMPULSE_ERROR_WARNING;
    desc = impulse_unistring_create_utf("Unexpected signal received", 8);
    break;
  }
  
#if HAVE_SIGINFO
  if (!msg)
  {
    switch(info->si_code)
    {
     case SI_USER:
      msg = impulse_unistring_create_utf("Killed by user", 8);
      break;
     case SI_QUEUE:
      msg = impulse_unistring_create_utf("Queued by user", 8);
      break;
     case SI_TIMER:
      msg = impulse_unistring_create_utf("Killed by timer expiration", 8);
      break;
     case SI_ASYNCIO:
      msg = impulse_unistring_create_utf("Killed by asynchronous IO notification", 8);
      break;
     case SI_MESGQ:
      msg = impulse_unistring_create_utf("Killed by arrival of a message", 8);
      break;
    }
  }
#endif
  
  impulse_error_raise(level, desc, msg);
  
  if (desc)
    impulse_unistring_destroy(desc);
  if (msg)
    impulse_unistring_destroy(msg);
}


#if HAVE_SIGINFO
static void impulse_signal(int signum, void (*signal_func)(int,
							   siginfo_t *,
							   void *))
#else
static void impulse_signal(int signum, void (*signal_func)(int))
#endif
{
  struct sigaction action;
  struct sigaction oldaction;

  sigfillset(&action.sa_mask);
  sigdelset(&action.sa_mask, SIGCHLD);
#if HAVE_SIGINFO
  action.sa_sigaction = signal_func;
  action.sa_flags     = SA_SIGINFO;
#else
  action.sa_handler   = signal_func;
  action.sa_flags     = 0;
#endif
  
  /* 
   * One can't help wondering what will happen when (if) this field is
   * implemented...
   */
  /* action.sa_restorer = NULL; */

  sigaction(signum, &action, &oldaction);
}


/*                            ======--------======                           */


static impulse_bool quit = 0;

static impulse_unistring opt_help = NULL;
static impulse_unistring opt_verbose = NULL;
static impulse_unistring opt_version = NULL;


impulse_int impulse_version(void)
{
  return (0 << 16) | (1 << 8) | 0;
}


void impulse_quit(void)
{
  quit = 1;
}

impulse_bool impulse_is_quit(void)
{
  return quit;
}


void impulse_main(void)
{
  while (!quit)
    impulse_event_handle();
}


static void command_cleaner(void *data)
{
  impulse_command_details details = (impulse_command_details)data;
  
  if (details->values)
  {
    impulse_int i;
    for (i = 0 ; i < details->count ; i++)
    {
      if (details->type == IMPULSE_COMMAND_TYPE_STRING)
	impulse_unistring_destroy(details->values[i].s);
    }
    impulse_free(details->values);
  }
  
  impulse_unistring_destroy(details->lopt);
  impulse_unistring_destroy(details->help);
  if (details->arg)
    impulse_unistring_destroy(details->arg);
  impulse_free(details);
}


static void command_hook(void)
{
  if (impulse_command_set(opt_help,NULL))
    impulse_command_help();
  if (impulse_command_set(opt_verbose, NULL))
    impulse_error_handle(IMPULSE_ERROR_INFO, IMPULSE_ERROR_PRINT, NULL);
  if (impulse_command_set(opt_version, NULL))
    ;
}


void impulse_init(impulse_uint argc,
		  impulse_char *argv[])
{
  impulse_int i;

#if HAVE_DEBUG_MCHECK
  if (getenv("IMPULSE_MCHECK"))
    impulse_mcheck();
#endif

  /* First off lets put in the signal handlers */
  impulse_signal(SIGHUP, _p_sighandler);
  impulse_signal(SIGINT, _p_sighandler);
  impulse_signal(SIGQUIT, _p_sighandler);
  impulse_signal(SIGILL, _p_sighandler);
  impulse_signal(SIGABRT, _p_sighandler);
  impulse_signal(SIGFPE, _p_sighandler);
  impulse_signal(SIGSEGV, _p_sighandler);
  impulse_signal(SIGPIPE, _p_sighandler);
  impulse_signal(SIGALRM, _p_sighandler);
  impulse_signal(SIGTERM, _p_sighandler);
  impulse_signal(SIGBUS, _p_sighandler);

  /* Now lets create the command line opts */
#if USE_THREADS
  impulse_mutex_lock(&command_opts_lock);
#endif
  uniargc = argc;
  uniargv = impulse_alloc(sizeof(impulse_unistring)*uniargc);
  for (i = 0 ; i < uniargc ; i++)
    uniargv[i] = impulse_unistring_create_utf(argv[i], 8);

  /*
   * Fix Me:
   * Do something about this nasty hack
   * See impulse_error.c impulse_module.c
   */
  prog_path = impulse_unistring_create(uniargv[0]);
  prog_name = NULL;
  
  command_lopts = impulse_hash_create();
  command_sopts = impulse_sparse_create();
  impulse_hash_cleaner(command_lopts, command_cleaner);
#if USE_THREADS
  impulse_mutex_unlock(&command_opts_lock);
#endif
  
#if USE_THREADS
  impulse_mutex_lock(&command_hooks_lock);
#endif
  
  command_hooks = impulse_list_create();
  impulse_list_append(command_hooks, command_hook);
  
#if USE_THREADS
  impulse_mutex_unlock(&command_hooks_lock);
#endif
  {
    impulse_unistring help = NULL;
    
    opt_help = impulse_unistring_create_utf("help", 8);
    help = impulse_unistring_create_utf("display this help and exit",8);
    impulse_command_option(opt_help, 0, help);
    impulse_unistring_destroy(help);
    
    opt_verbose = impulse_unistring_create_utf("verbose", 8);
    help = impulse_unistring_create_utf("display verbose information messages",8);
    impulse_command_option(opt_verbose, 0, help);
    impulse_unistring_destroy(help);
    
    opt_version = impulse_unistring_create_utf("version", 8);
    help = impulse_unistring_create_utf("output version information and exit",8);
    impulse_command_option(opt_version, 0, help);
    impulse_unistring_destroy(help);
  }
}


void impulse_cleanup(void)
{
  impulse_int i;
#if USE_THREADS
  impulse_mutex_lock(&command_hooks_lock);
#endif
  
  impulse_list_destroy(command_hooks);
  
#if USE_THREADS
  impulse_mutex_unlock(&command_hooks_lock);
#endif
  
#if USE_THREADS
  impulse_mutex_lock(&command_opts_lock);
#endif
  for (i = 0 ; i < uniargc ; i++)
    impulse_unistring_destroy(uniargv[i]);
  impulse_free(uniargv);
  uniargv = NULL;
  uniargc = 0;
  impulse_hash_destroy(command_lopts);
  impulse_sparse_destroy(command_sopts);
#if USE_THREADS
  impulse_mutex_unlock(&command_opts_lock);
#endif
}


void impulse_command_help(void)
{
  impulse_iterator iterator = NULL;
  impulse_char *str = NULL;
  
#if USE_THREADS
  impulse_mutex_lock(&command_opts_lock);
#endif
  
  str = impulse_unistring_to_utf(prog_path, 8, 0);
  printf("syntax: %s [options]\n", str);
  impulse_free(str);
  
  iterator = impulse_hash_elements(command_lopts);
  while (!impulse_iterator_eof(iterator))
  {
    impulse_char *lopt, *arg = NULL, *help;
    impulse_command_details details = (impulse_command_details)impulse_iterator_next(iterator);
    impulse_char pad[21] = "                    ";

    lopt = impulse_unistring_to_utf(details->lopt, 8, 0);
    if (details->arg)
      arg = impulse_unistring_to_utf(details->arg, 8, 0);
    help = impulse_unistring_to_utf(details->help, 8, 0);

    if (arg)
    {
      impulse_int padn = 20 - (strlen(lopt) + strlen(arg) + 1);
      pad[(padn < 0) ? 0 : padn] = '\0';
    }
    else
    {
      impulse_int padn = 20 - strlen(lopt);
      pad[(padn < 0) ? 0 : padn] = '\0';
    }
    if (details->sopt)
    {
      if (arg)
	printf("-%c, --%s=%s%s%s\n", (char)details->sopt, lopt, arg, pad, help);
      else
	printf("-%c, --%s%s%s\n", (char)details->sopt, lopt, pad, help);
    }
    else
    {
      if (arg)
	printf("    --%s=%s%s%s\n", lopt, arg, pad, help);
      else
	printf("    --%s%s%s\n", lopt, pad, help);
    }
    impulse_free(help);
    if (arg)
      impulse_free(arg);
    impulse_free(lopt);
  }
  impulse_iterator_destroy(iterator);
#if USE_THREADS
  impulse_mutex_unlock(&command_opts_lock);
#endif
  exit(0);
}


static void command_parse_val(impulse_unistring val,
			      impulse_command_details details)
{
  if (((details->type == IMPULSE_COMMAND_TYPE_INTEGER) ||
       (details->type == IMPULSE_COMMAND_TYPE_DECIMAL) ||
       (details->type == IMPULSE_COMMAND_TYPE_STRING)) &&
      details->count)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Repeated command line arguments", 8);
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, details->lopt);
    impulse_unistring_destroy(desc);
  }
  details->values = impulse_realloc(details->values, sizeof(union _impulse_command_value) * (details->count+1));
  
  switch (details->type)
  {
   case IMPULSE_COMMAND_TYPE_INTEGER:
   case IMPULSE_COMMAND_TYPE_INTEGERS:
    {
      impulse_char *str = impulse_unistring_to_utf(val, 8, 0);
      details->values[details->count].i = atoi(str);
      impulse_free(str);
    }
    break;
    
   case IMPULSE_COMMAND_TYPE_DECIMAL:
   case IMPULSE_COMMAND_TYPE_DECIMALS:
    {
      impulse_char *str = impulse_unistring_to_utf(val, 8, 0);
      details->values[details->count].d = atof(str);
      impulse_free(str);
    }
    break;
	  
   case IMPULSE_COMMAND_TYPE_STRING:
   case IMPULSE_COMMAND_TYPE_STRINGS:
    details->values[details->count].s = impulse_unistring_create(val);
    break;
    
   default:
    break;
  }
  details->count++;
}


impulse_uint impulse_command_parse(void)
{
  impulse_int i;
  impulse_int end = 0;
  
#if USE_THREADS
  impulse_mutex_lock(&command_opts_lock);
#endif
  
  for (i = 1 ; i < uniargc ; i++)
  {
    enum { ARG_LONG, ARG_SHORT, ARG_VAL } type;

    if (uniargv[i]->length < 2)
    {
      impulse_unistring desc = impulse_unistring_create_utf("Malformed command line argument", 8);
      impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, uniargv[i]);
      impulse_unistring_destroy(desc);
    }

    if (uniargv[i]->data[0] != (impulse_unichar)'-')
      type = ARG_VAL;
    else
    {
      if (uniargv[i]->data[1] != (impulse_unichar)'-')
	type = ARG_SHORT;
      else
	type = ARG_LONG;
    }
    
    if ((type == ARG_LONG) &&
	(uniargv[i]->length == 2))
    {
      lastarg = i;
      break;
    }
    
    if (type == ARG_VAL)
    {
      if (!end)
	end = i;
      continue;
    }

    if (end)
    {
      impulse_unistring desc = impulse_unistring_create_utf("Misplaced command line argument", 8);
      impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, uniargv[i]);
      impulse_unistring_destroy(desc);
    } 

    if (type == ARG_LONG)
    {
      impulse_int offset = impulse_unistring_find_chr(uniargv[i], (impulse_unichar)'=');
      /* At this point we must have a long argument */
      impulse_unistring arg = impulse_unistring_create_empty();
      impulse_unistring val = NULL;
      impulse_command_details details = NULL;
      
      if (offset < 0)
	impulse_unistring_substring(arg, uniargv[i], 2, uniargv[i]->length);
      else
      {
	impulse_unistring_substring(arg, uniargv[i], 2, offset);
	val = impulse_unistring_create_empty();
	impulse_unistring_substring(val, uniargv[i], offset+1, uniargv[i]->length);
      }
      
      details = impulse_hash_recover(command_lopts, arg);
      if (!details)
      {
	impulse_unistring desc = impulse_unistring_create_utf("Unknown command line argument", 8);
	impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, uniargv[i]);
      }
      
      if (details->type == IMPULSE_COMMAND_TYPE_NONE)
      {
	if (val)
	{
	  impulse_unistring desc = impulse_unistring_create_utf("Malformed command line argument", 8);
	  impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, uniargv[i]);
	}
	details->count++;
      }
      else
      {
	if (!val)
	{
	  if (++i == uniargc)
	  {
	    impulse_unistring desc = impulse_unistring_create_utf("Malformed command line argument", 8);
	    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, uniargv[i]);
	  }
	  val = impulse_unistring_create(uniargv[i]);
	}
	command_parse_val(val, details);
      }
      if (val)
	impulse_unistring_destroy(val);
      if (arg)
	impulse_unistring_destroy(arg);
    }
    else
    {
      impulse_int offset = impulse_unistring_find_chr(uniargv[i], (impulse_unichar)'=');
      impulse_int j;
      
      if ((offset != -1) && (offset != 2))
      {
	impulse_unistring desc = impulse_unistring_create_utf("Malformed command line argument", 8);
	impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, uniargv[i]);
      }
      
      for (j = 1 ; j < uniargv[i]->length ; j++)
      {
	impulse_unichar opt = uniargv[i]->data[j];
	impulse_command_details details = impulse_sparse_recover(command_sopts, opt);
	impulse_unistring val = NULL;

	if (!details)
	{
	  impulse_unistring desc = impulse_unistring_create_utf("Unknown command line argument", 8);
	  impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, uniargv[i]);
	}
	
	if (details->type == IMPULSE_COMMAND_TYPE_NONE)
	{
	  if (offset != -1)
	  {
	    impulse_unistring desc = impulse_unistring_create_utf("Malformed command line argument", 8);
	    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, uniargv[i]);
	  }
	  details->count++;
	}
	else
	{
	  impulse_unistring val = impulse_unistring_create_empty();
	  if (offset != -1)
	    impulse_unistring_substring(val, uniargv[i], offset+1, uniargv[i]->length);
	  else
	  {
	    if (++i == uniargc)
	    {
	      impulse_unistring desc = impulse_unistring_create_utf("Malformed command line argument", 8);
	      impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, uniargv[i]);
	    }
	    val = impulse_unistring_create(uniargv[i]);
	  }
	  command_parse_val(val, details);
	}
	if (val)
	  impulse_unistring_destroy(val);
      }
    }
  }

#if USE_THREADS
  impulse_mutex_unlock(&command_opts_lock);
#endif

#if USE_THREADS
  impulse_mutex_lock(&command_hooks_lock);
#endif

  if (!impulse_list_empty(command_hooks))
  {
    impulse_iterator iterator = impulse_list_elements(command_hooks);
    while (!impulse_iterator_eof(iterator))
    {
      impulse_command_func func = (impulse_command_func)impulse_iterator_next(iterator);
      
      (*func)();
    }
    impulse_iterator_destroy(iterator);
  }

#if USE_THREADS
  impulse_mutex_unlock(&command_hooks_lock);
#endif
  return 0;
}


void impulse_command_option(impulse_unistring lopt,
			    impulse_unichar sopt,
			    impulse_unistring help)
{
  impulse_command_option_arg(lopt, sopt, NULL, IMPULSE_COMMAND_TYPE_NONE, help);
}


void impulse_command_option_arg(impulse_unistring lopt,
				impulse_unichar sopt,
				impulse_unistring arg,
				impulse_command_type type,
				impulse_unistring help)
{
  impulse_command_details details = NULL;
#if USE_THREADS
  impulse_mutex_lock(&command_opts_lock);
#endif
  
  details = impulse_alloc(sizeof(struct _impulse_command_details));
  details->lopt = impulse_unistring_create(lopt);
  details->sopt = sopt;
  details->type = type;
  if (arg)
    details->arg = impulse_unistring_create(arg);
  else
    details->arg = NULL;
  details->help = impulse_unistring_create(help);
  details->count = 0;
  details->values = NULL;
  
  impulse_hash_store(command_lopts, details->lopt, details);
  if (sopt)
    impulse_sparse_add(command_sopts, sopt, details);
  
#if USE_THREADS
  impulse_mutex_unlock(&command_opts_lock);
#endif
}


impulse_uint impulse_command_set(impulse_unistring lopt,
				 impulse_command_value *values)
{
  impulse_command_details details;
  impulse_uint count = 0;
#if USE_THREADS
  impulse_mutex_lock(&command_opts_lock);
#endif
  
  details = (impulse_command_details)impulse_hash_recover(command_lopts, lopt);
  if (!details)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Unknown command line option", 8);
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, lopt);
  }

  count = details->count;
  if (values)
  {
    if (count)
      *values = details->values;
    else
      *values = NULL;
  }
  
#if USE_THREADS
  impulse_mutex_unlock(&command_opts_lock);
#endif
  return count;
}


void impulse_command_hook(impulse_command_func func)
{
#if USE_THREADS
  impulse_mutex_lock(&command_hooks_lock);
#endif
  
  impulse_list_append(command_hooks, func);
  
#if USE_THREADS
  impulse_mutex_unlock(&command_hooks_lock);
#endif
}


impulse_uint impulse_command_args(impulse_uint *argc,
				  impulse_unistring **argv)
{
  *argv = uniargv;
  *argc = uniargc;
  return lastarg;
}


/* End of file */
