/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * module.c
 * --------
 *
 * Impulse runtime dynamic library loading
 */

/*********************************************************************
 * 
 * VIW (Very Important Warning)
 * 
 * No functions used impulse_module_* must *EVER* call
 * 
 *  impulse_unistring_to_8bit
 *  impulse_unistring_from_8bit
 *  impulse_text_get
 * 
 * All of these functions make use of dynamically loaded modules,
 * and thus would have huge potential for infinite loops.....
 * 
 * Unfortunately this means that error messages in this module
 * can not be internationalised. In fact do we really need error 
 * messages here, seeing as the caller will have to deal with the
 * NULL return values and will likely display a suitable message
 * itself.
 * 
 ********************************************************************/

#include "../../../config.h"

#include "impulse/util/module.h"
#include "impulse/util/error.h"
#include "impulse/util/memory.h"
#include "impulse/util/hash.h"
#include "impulse/util/thread.h"

#if HAVE_MODULE_DL
# include <dlfcn.h>
#elif HAVE_MODULE_DLD
# include <dld.h>
#elif HAVE_MODULE_WIN32
# include <windows.h>
#elif HAVE_MODULE_SHL
# include <dl.h>
#else
#error "Doh, dynamic modules don't seem to be supported on your platform"
#endif


struct _impulse_module
{
  impulse_unistring name;
  impulse_unistring file;
  void *handle;
  impulse_int refs;
  
  impulse_int ndeps;
  struct _impulse_module **deps;
};


static impulse_hash modules = NULL;
#if USE_THREADS
static impulse_mutex_static module_mutex = IMPULSE_MUTEX_CREATE_STATIC;
#endif


void _p_impulse_module_error(impulse_unistring desc)
{
#if HAVE_MODULE_SHL
  impulse_error_syscall(IMPULSE_ERROR_WARNING, desc);
#else
# if HAVE_MODULE_DL
#  if HAVE_MODULE_DL_DLERROR
  impulse_unistring msg = impulse_unistring_create_utf(dlerror(), 8);
#  else
  impulse_unistring msg = impulse_unistring_create_utf("Unknown error", 8);
#  endif
# elif HAVE_MODULE_DLD
  impulse_unistring msg = impulse_unistring_create_utf(dld_strerror(dld_errno), 8);
# elif HAVE_MODULE_WIN32
  impulse_unistring msg = impulse_unistring_create_utf("Unknown error", 8);
# endif
  impulse_error_raise(IMPULSE_ERROR_WARNING, desc, msg);
  impulse_unistring_destroy(msg);
#endif
}


impulse_module _p_impulse_module_create(impulse_unistring name)
{
  impulse_module mod = impulse_alloc(sizeof(struct _impulse_module));
  int offset = -1;
  char *cfile = NULL;
  
  mod->name = impulse_unistring_create_empty();
  mod->file = impulse_unistring_create_empty();
  mod->handle = NULL;
  mod->refs = 1;
  mod->ndeps = 0;
  mod->deps = NULL;
  
  
  offset = impulse_unistring_rfind_chr(name, (impulse_unichar)'/');
  if (offset == -1)
    impulse_unistring_assign(mod->name, name);
  else
    impulse_unistring_substring(mod->name, name, offset+1, name->length);

  impulse_unistring_assign(mod->file, name);
  impulse_unistring_append_utf(mod->file, ".so", 8);
  cfile = impulse_unistring_to_utf(mod->file, 8, 0);
  
#if HAVE_MODULE_DL
  mod->handle = dlopen(cfile, RTLD_NOW | RTLD_GLOBAL);
#elif HAVE_MODULE_DLD
  mod->handle = dld_link(cfile);
#elif HAVE_MODULE_WIN32
  mod->handle = LoadLibrary(cfile);
#elif HAVE_MODULE_SHL
  mod->handle = shl_load(cfile, BIND_IMMEDIATE | BIND_NOSTART, 0L);
#endif
  
  impulse_free(cfile);
  
  if (!mod->handle)
  {
#if 0
    impulse_unistring desc = impulse_unistring_create_utf("Cannot load dynamic module", 8);
    _p_impulse_module_error(desc);
    impulse_unistring_destroy(desc);
#endif
    impulse_unistring_destroy(mod->name);
    impulse_unistring_destroy(mod->file);
    impulse_free(mod);
    return NULL;
  }
  
  return mod;
}


void _p_impulse_module_destroy(impulse_module module)
{
  int i;
  
#if HAVE_MODULE_DL
  dlclose(module->handle);
#elif HAVE_MODULE_DLD
  impulse_char *cstr = impulse_unistring_to_utf(module->file, 8, 0);
  dld_unlink_by_file(cstr, 1);
  impulse_free(cstr);
#elif HAVE_MODULE_WIN32
  FreeLibrary(module->handle);
#elif HAVE_MODULE_SHL
  shl_unload(module->handle);
#endif
  
  impulse_unistring_destroy(module->name);
  impulse_unistring_destroy(module->file);
  for (i = 0 ; i < module->ndeps ; i++)
    impulse_module_destroy(module->deps[i]);
  
  impulse_free(module);
}


void *_p_impulse_module_symbol(impulse_module module,
                               impulse_unistring name)
{
  impulse_char *cstr = impulse_unistring_to_utf(name, 8, 0);
  void *sym = NULL;
  
#if HAVE_MODULE_DL
  sym = dlsym(module->handle, cstr);
#elif HAVE_MODULE_DLD
  sym = dld_get_func(cstr);
#elif HAVE_MODULE_WIN32
  sym = GetProcAddress(module->handle, cstr);
#elif HAVE_MODULE_SHL
  if (shl_findsym(module->handle, cstr, TYPE_UNDEFINED, &sym) < 0)
    sym = NULL;
#endif
  impulse_free(cstr);
  
  if (!sym)
  {
    impulse_unistring pref = impulse_unistring_create_utf("_", 8);

    impulse_unistring_prepend(pref, module->name);
    impulse_unistring_append(pref, name);
    cstr = impulse_unistring_to_utf(pref, 8, 0);

#if HAVE_MODULE_DL
    sym = dlsym(module->handle, cstr);
#elif HAVE_MODULE_DLD
    sym = dld_get_func(cstr);
#elif HAVE_MODULE_WIN32
    sym = GetProcAddress(module->handle, cstr);
#elif HAVE_MODULE_SHL
    if (shl_findsym(module->handle, cstr, TYPE_UNDEFINED, &sym) < 0)
      sym = NULL;
#endif
    impulse_free(cstr);
    impulse_unistring_destroy(pref);
    
    if (!sym)
    {
#if 0
      impulse_unistring desc = impulse_unistring_create_utf("Cannot find symbol in dynamic module", 8);
      _p_impulse_module_error(desc);
      impulse_unistring_destroy(desc);
#endif
    }
  }
  
  return sym;
}

#if HAVE_MODULE_DLD
extern impulse_unistring prog_path;
#endif

impulse_module impulse_module_create(impulse_unistring name)
{
  impulse_module mod = NULL;

#if USE_THREADS
  impulse_mutex_lock(&module_mutex);
#endif
  
  if (!modules)
  {
    modules = impulse_hash_create();
#if HAVE_MODULE_DLD
    {
      impulse_char *cstr = impulse_unistring_to_utf(prog_path, 8, 0);
      dld_init(cstr);
      impulse_free(cstr);
    }
#endif
  }
  
  if (!(mod = impulse_hash_recover(modules, name)))
  {
    mod = _p_impulse_module_create(name);
    if (mod)
      impulse_hash_store(modules, mod->file, mod);
  }
  else
  {
    mod->refs++;
  }
  
#if USE_THREADS
  impulse_mutex_unlock(&module_mutex);
#endif
  return mod;
}


/*
 * Fix Me:
 * Instead of releasing the module the moment its reference
 * count goes to zero, we could do something clever:
 * Place the module in a 'delayed release' list and set a timer
 * to fully release it in say 30 seconds time or so. This will 
 * improve performance if a module is needed just a short time
 * after it is released.
 */
void impulse_module_destroy(impulse_module module)
{
#if USE_THREADS
  impulse_mutex_lock(&module_mutex);
#endif
  
  if (!(--(module->refs)))
  {
    impulse_hash_store(modules, module->file, NULL);
    _p_impulse_module_destroy(module);
  }

#if USE_THREADS
  impulse_mutex_unlock(&module_mutex);
#endif
}


void *impulse_module_symbol(impulse_module module,
			    impulse_unistring name)
{
  void *sym = NULL;
#if USE_THREADS
  impulse_mutex_lock(&module_mutex);
#endif
  sym = _p_impulse_module_symbol(module, name);
  
#if USE_THREADS
  impulse_mutex_unlock(&module_mutex);
#endif
  return sym;
}


/* End of file */
