/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * queue.c
 * -------
 *
 * Double headed queues
 */

#include "../../../config.h"

#include "impulse/util/memory.h"
#include "impulse/util/queue.h"

/*                            ======--------======                           */

struct _impulse_queue_bucket
{
  void *val;
  struct _impulse_queue_bucket *next;
  struct _impulse_queue_bucket *prev;
};


typedef struct _impulse_queue_bucket *impulse_queue_bucket;

struct _impulse_queue
{
  impulse_queue_bucket head;
  impulse_queue_bucket tail;
  
  impulse_uint count;
  
  impulse_queue_cleaner_func cleaner;
};


void impulse_queue_iterator_destroy(void *data);
impulse_bool impulse_queue_iterator_eof(void *data);
void impulse_queue_iterator_reset(void *data);
void *impulse_queue_iterator_next(void *data);

static struct _impulse_iterator_funcs iterator_funcs = 
{
  impulse_queue_iterator_destroy,
  impulse_queue_iterator_eof,
  impulse_queue_iterator_reset,
  impulse_queue_iterator_next
};

/*                            ======--------======                           */

impulse_queue impulse_queue_create(void)
{
  impulse_queue queue = impulse_alloc(sizeof(struct _impulse_queue));
  
  queue->head = queue->tail = NULL;
  queue->count = 0;
  queue->cleaner = NULL;
  
  return queue;
}


void impulse_queue_destroy(impulse_queue queue)
{
  impulse_queue_clear(queue);
  impulse_free(queue);
}


void impulse_queue_cleaner(impulse_queue queue,
			   impulse_queue_cleaner_func func)
{
  queue->cleaner = func;
}


impulse_iterator impulse_queue_elements(impulse_queue queue)
{
  impulse_iterator iterator = impulse_iterator_create(&iterator_funcs, NULL);
  return iterator;
}


impulse_bool impulse_queue_empty(impulse_queue queue)
{
  return queue->count == 0;
}


impulse_uint impulse_queue_count(impulse_queue queue)
{
  return queue->count;
}


void impulse_queue_enqueue_end(impulse_queue queue,
			       void *data,
			       impulse_bool head)
{
  impulse_queue_bucket temp = impulse_alloc(sizeof(struct _impulse_queue_bucket));

  if (!queue->head)
  {
    queue->head = temp;
    queue->tail = temp;
  }
  else if (head)
  {
    queue->head->prev = temp;
    temp->next = queue->head;
    queue->head = temp;
  }
  else
  {
    temp->prev = queue->tail;
    queue->tail->next = temp;
    queue->tail = temp;
  }
  queue->count++;
}


void *impulse_queue_serve_end(impulse_queue queue,
			      impulse_bool head)
{
  impulse_queue_bucket temp = NULL;
  void *data = NULL;
  
  if (!queue->count)
    return NULL;
  
  if (head)
  {
    temp = queue->head;
    queue->head = queue->head->next;
    if (queue->head)
      queue->head->prev = NULL;
    else
      queue->tail = NULL;
  }
  else
  {
    temp = queue->tail;
    queue->tail = queue->tail->prev;
    if (queue->tail)
      queue->tail->next = NULL;
    else
      queue->head = NULL;
  }
  queue->count--;
  data = temp->val;
  impulse_free(temp);
  return data;
}


void impulse_queue_remove(impulse_queue queue,
			  void *data)
{
  impulse_queue_bucket temp = queue->head;
  if (!queue->count)
    return;
  
  while (temp)
  {
    if (temp->val == data)
    {
      if (temp->prev)
        temp->prev->next = temp->next;
      if (temp->next)
        temp->next->prev = temp->prev;
      
      if (temp == queue->head)
        queue->head = temp->next;
      if (temp == queue->tail)
        queue->tail = temp->prev;
      
      impulse_free(temp);
      queue->count--;
      return;
    }
    temp = temp->next;
  }
}


void impulse_queue_clear(impulse_queue queue)
{
  impulse_queue_bucket temp = queue->head, next = NULL;
  while (temp)
  {
    next = temp->next;
    if (queue->cleaner)
      (*queue->cleaner)(temp->val);
    
    impulse_free(temp);
    temp = next;
  }
  queue->head = NULL;
  queue->tail = NULL;
  queue->count = 0;
}


impulse_bool impulse_queue_contains(impulse_queue queue,
				    void *data)
{
  impulse_queue_bucket temp = queue->head;
  while (temp)
  {
    if (temp->val == data)
      return 1;
    temp = temp->next;
  }
  return 0;
}


void impulse_queue_iterator_destroy(void *data)
{
}


impulse_bool impulse_queue_iterator_eof(void *data)
{
  return 1;
}


void impulse_queue_iterator_reset(void *data)
{
}


void *impulse_queue_iterator_next(void *data)
{
  return NULL;
}


/* End of file */
