/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 */

/*
 * sparse.c
 * --------
 *
 * Sparse arrays
 */

/*
 * This implementation uses n-ary trees for storing the nodes.
 * In this case n is related to the granularity by the simple
 * formula n=(1<<granularity)
 */

#include "../../../config.h"

#include "impulse/util/error.h"
#include "impulse/util/memory.h"
#include "impulse/util/sparse.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*                            ======--------======                           */

enum _p_impulse_sparse_item_usage {
  _p_impulse_sparse_item_UNUSED,
  _p_impulse_sparse_item_FILLED,
  _p_impulse_sparse_item_MORE
};

struct _impulse_sparse_item
{
  enum _p_impulse_sparse_item_usage usage;
  impulse_uint key;

  struct _impulse_sparse_item *parent;
  union {
    void *data;
    struct _impulse_sparse_item *child;
  } contents;
};

struct _impulse_sparse_iter_item
{
  struct _impulse_sparse_iterator *data;
  struct _impulse_sparse_iter_item *next;
};

typedef struct _impulse_sparse_item impulse_sparse_item;


struct _impulse_sparse {
  impulse_sparse_item *child;
  struct _impulse_sparse_iter_item *iters;
  
  impulse_uint granularity;
  impulse_uint count;
  
  impulse_sparse_cleaner_func cleaner;
};


struct _impulse_sparse_iterator
{
  impulse_sparse array;
  impulse_sparse_item current;
  impulse_uint cell;
  impulse_bool keys;
};


typedef struct _impulse_sparse_iterator *impulse_sparse_iterator;

impulse_sparse_iterator impulse_sparse_iterator_create(impulse_sparse table,
                                                   impulse_bool keys);
void impulse_sparse_iterator_destroy(void *data);
impulse_bool impulse_sparse_iterator_eof(void *data);
void impulse_sparse_iterator_reset(void *data);
void *impulse_sparse_iterator_next(void *data);

static struct _impulse_iterator_funcs iterator_funcs = 
{
  impulse_sparse_iterator_destroy,
  impulse_sparse_iterator_eof,
  impulse_sparse_iterator_reset,
  impulse_sparse_iterator_next
};


static void _p_impulse_sparse_error(impulse_unistring msg)
{
  impulse_unistring desc = impulse_unistring_create_utf("Sparse array has become corrupted", 8);
  impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, msg);
  impulse_unistring_destroy(desc);
}


impulse_sparse impulse_sparse_create(void)
{
  return impulse_sparse_create_gran(2);
}


impulse_sparse impulse_sparse_create_gran(impulse_uint granularity)
{
  impulse_sparse array = impulse_alloc(sizeof(struct _impulse_sparse));
  impulse_uint x;

  array->granularity = granularity;
  array->count = 0;
  array->cleaner = NULL;
  
  array->child = impulse_alloc(sizeof(struct _impulse_sparse) * (1 << array->granularity));
  for (x = 0 ; x < (1 << array->granularity) ; x++)
  {
    array->child[x].usage = _p_impulse_sparse_item_UNUSED;
    array->child[x].key = 0;
    array->child[x].parent = NULL;
    array->child[x].contents.data = NULL;
    array->child[x].contents.child = NULL;
  }

  return array;
}


void impulse_sparse_destroy(impulse_sparse array)
{
  impulse_sparse_clear(array);
  impulse_free(array->child);
  impulse_free(array);
}


void impulse_sparse_cleaner(impulse_sparse array,
			    impulse_sparse_cleaner_func func)
{
  array->cleaner = func;
}


impulse_iterator impulse_sparse_elements(impulse_sparse array)
{
  impulse_sparse_iterator it = impulse_sparse_iterator_create(array, 0);
  impulse_iterator iterator = impulse_iterator_create(&iterator_funcs,
                                                      it);
  return iterator;
}


impulse_iterator impulse_sparse_keys(impulse_sparse array)
{
  impulse_sparse_iterator it = impulse_sparse_iterator_create(array, 0);
  impulse_iterator iterator = impulse_iterator_create(&iterator_funcs,
                                                      it);
  return iterator;
}


impulse_bool impulse_sparse_empty(impulse_sparse array)
{
  return array->count == 0;
}


impulse_uint impulse_sparse_count(impulse_sparse array)
{
  return array->count;
}


void impulse_sparse_add(impulse_sparse array,
			impulse_uint key,
			void *data)
{
  impulse_sparse_item *item;
  impulse_uint level;
  impulse_uint x;
  impulse_uint maxlevel = 32 / array->granularity;
  impulse_uint mask = (1 << array->granularity) - 1;
  impulse_uint num[maxlevel];

  for (level = 0 ; level < maxlevel ; level++)
    num[level] = (key >> (level * array->granularity)) & mask;

  /* Find a free impulse_sparse_item */
  item = &array->child[num[0]];
  level = 1;

  while (item->usage != _p_impulse_sparse_item_UNUSED && item->key != key)
  {
    if (level >= maxlevel)
    {
      impulse_unistring msg = impulse_unistring_create_utf("Sparse item mis-numbered", 8);
      _p_impulse_sparse_error(msg);
      impulse_unistring_destroy(msg);
    }

    if (item->usage == _p_impulse_sparse_item_MORE)
    {
      item = &item->contents.child[num[level]];
      level++;
    }
    else
    {
      int other_hash;
      void *old_data;
      
      /* Presumably filled, so we need to expand down a level */
      
      /* Create the new array */
      old_data = item->contents.data;
      item->contents.child = impulse_alloc(sizeof(struct _impulse_sparse_item) * (1 << array->granularity));
      
      /* Fill it with empties */
      for (x = 0 ; x < (1 << array->granularity) ; x++)
      {
	item->contents.child[x].usage = _p_impulse_sparse_item_UNUSED;
	item->contents.child[x].key = 0;
	item->contents.child[x].parent = item;
	item->contents.child[x].contents.data = NULL;
	item->contents.child[x].contents.child = NULL;
      }
      
      /* Move the old item into the new array */
      other_hash = (item->key >> (array->granularity * level))&mask;
    
      item->contents.child[other_hash].usage = _p_impulse_sparse_item_FILLED;
      item->contents.child[other_hash].key = item->key;
      item->contents.child[other_hash].contents.data = old_data;

      /* Mark this item as an array node */
      item->usage = _p_impulse_sparse_item_MORE;
      
      /* The new item value */
      item = &item->contents.child[num[level]];
    
      level++;
    }
  }
  
  item->usage = _p_impulse_sparse_item_FILLED;
  item->key = key;
  item->contents.data = data;
  
  array->count++;
}


void *impulse_sparse_recover(impulse_sparse array,
			     impulse_uint key)
{
  struct _impulse_sparse_item *item;
  impulse_uint maxlevel = 32 / array->granularity;
  impulse_uint num[maxlevel];
  impulse_uint level;
  impulse_uint mask = (1 << array->granularity)-1;

  for (level = 0 ; level < maxlevel ; level++)
    num[level] = (key >> (level * array->granularity)) & mask;

  item = &array->child[num[0]];
  level = 1;

  while (item->usage == _p_impulse_sparse_item_MORE)
  {
    if (level >= maxlevel)
    {
      impulse_unistring msg = impulse_unistring_create_utf("Sparse item mis-numbered", 8);
      _p_impulse_sparse_error(msg);
      impulse_unistring_destroy(msg);
    }
    
    item = &item->contents.child[num[level]];
    level++;
  }

  if (item->usage == _p_impulse_sparse_item_UNUSED || item->key != key)
    return NULL;

  return item->contents.data;
}


void *impulse_sparse_remove(impulse_sparse array,
			    impulse_uint key)
{
  struct _impulse_sparse_item *item;
  impulse_uint maxlevel = 32 / array->granularity;
  impulse_uint num[maxlevel];
  impulse_uint level;
  impulse_uint mask = (1 << array->granularity)-1;
  void *data;

  for (level = 0 ; level < maxlevel ; level++)
    num[level] = (key >> (level * array->granularity)) & mask;

  item = &array->child[num[0]];
  level = 1;

  while (item->usage == _p_impulse_sparse_item_MORE)
  {
    if (level >= maxlevel)
    {
      impulse_unistring msg = impulse_unistring_create_utf("Sparse item mis-numbered", 8);
      _p_impulse_sparse_error(msg);
      impulse_unistring_destroy(msg);
    }
    
    item = &item->contents.child[num[level]];
    level++;
  }

  if (item->usage == _p_impulse_sparse_item_UNUSED || item->key != key)
    return NULL;

  data = item->contents.data;
  
  item->usage = _p_impulse_sparse_item_UNUSED;
  item->key = 0;
  item->contents.data = NULL;
  
  return data;
}


static void impulse_sparse_item_clear(impulse_sparse array,
				      impulse_sparse_item *item)
{
  if (item->usage == _p_impulse_sparse_item_MORE)
  {
    impulse_uint x;
    for (x = 0 ; x < (1 << array->granularity) ; x++)
    {
      impulse_sparse_item_clear(array, &item->contents.child[x]);
    }
    
    impulse_free(item->contents.child);
  } 
  else if (item->usage == _p_impulse_sparse_item_FILLED)
  {
    if (array->cleaner)
      (*array->cleaner)(item->contents.data);
  }
}


void impulse_sparse_clear(impulse_sparse array)
{
  impulse_uint x;
  
  for (x = 0 ; x < (1 << array->granularity) ; x++)
  {
    impulse_sparse_item_clear(array, &array->child[x]);
    
    array->child[x].usage = _p_impulse_sparse_item_UNUSED;
    array->child[x].key = 0;
    array->child[x].parent = NULL;
    array->child[x].contents.child = 0;
    array->child[x].contents.data = 0;
  }
  array->count = 0;
}


impulse_bool impulse_sparse_contains(impulse_sparse array,
				     void *data)
{
  return 0;
}


impulse_bool impulse_sparse_contains_key(impulse_sparse array,
					 impulse_uint key)
{
  return impulse_sparse_recover(array, key) != NULL;
}


impulse_sparse_iterator impulse_sparse_iterator_create(impulse_sparse table,
						       impulse_bool keys)
{
  impulse_sparse_iterator iterator = impulse_alloc(sizeof(struct _impulse_sparse_iterator));
  
  return iterator;
}


void impulse_sparse_iterator_destroy(void *data)
{
  impulse_free(data);
}


impulse_bool impulse_sparse_iterator_eof(void *data)
{
  return 1;
}


void impulse_sparse_iterator_reset(void *data)
{
}


void *impulse_sparse_iterator_next(void *data)
{
  return NULL;
}



/* End of file */
