/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * thread_pth.c
 * ------------
 *
 * Impulse threading routines
 * GNU Portable Threading Library implementation
 */

#include <errno.h>

/*                            ======--------======                           */

/* Thread function defintions */


void impulse_thread_init(void)
{
  pth_init();
}


void impulse_thread_cleanup(void)
{
}


impulse_thread impulse_thread_create(impulse_thread_main func,
				     void *argument)
{
  pth_attr_t attr = NULL;
  pth_t      thread;
  
  if (attr == NULL)
    attr = pth_attr_new();
  
  pth_attr_set(attr, PTH_ATTR_JOINABLE, TRUE);
  
  thread = pth_spawn(attr,
		     func,
		     argument);

  pth_attr_destroy(attr);

  return thread;
}


impulse_thread impulse_thread_create_detach(impulse_thread_main func,
					    void *argument)
{
  pth_attr_t attr = NULL;
  pth_t      thread;
  
  if (attr == NULL)
    attr = pth_attr_new();
  
  pth_attr_set(attr, PTH_ATTR_JOINABLE, FALSE);
  
  thread = pth_spawn(attr,
		     func,
		     argument);
  
  pth_attr_destroy(attr);
  
  return thread;
}


void impulse_thread_exit(void * status)
{
  pth_exit(status);
}


void impulse_thread_join(impulse_thread t, void **status)
{
  THREAD_ERROR(pth_join(t, status));
}


impulse_thread impulse_thread_self(void)
{
  return pth_self();
}

impulse_bool impulse_thread_same(impulse_thread a,
				 impulse_thread b)
{
  return a == b;
}


void impulse_thread_nap(impulse_int secs,
			impulse_int usecs)
{
  struct timeval wakeup;

  wakeup.tv_sec  = secs;
  wakeup.tv_usec = usecs;
  THREAD_ERROR(pth_nap(wakeup));
}


void impulse_thread_yield(void)
{
  THREAD_ERROR(pth_yield(NULL));
}


/*                            ======--------======                           */

/* Mutex functions */

impulse_mutex impulse_mutex_create(void)
{
  pth_mutex_t *mutex;

  mutex = impulse_alloc(sizeof(pth_mutex_t));
  THREAD_ERROR(pth_mutex_init(mutex));

  return mutex;
}

impulse_mutex impulse_mutex_create_recursive(void)
{
  pth_mutex_t *mutex;

  mutex = impulse_alloc(sizeof(pth_mutex_t));
  THREAD_ERROR(pth_mutex_init(mutex));

  return mutex;
}

void impulse_mutex_lock(impulse_mutex mutex)
{
  THREAD_ERROR(pth_mutex_acquire(mutex, FALSE, NULL));
}

impulse_bool impulse_mutex_trylock(impulse_mutex mutex)
{
  impulse_int ret;
  ret = pth_mutex_acquire(mutex, TRUE, NULL);
  if (ret == EBUSY)
    return 0;
  else if (ret < 0)
    THREAD_ERROR(ret);
  return 1;
}

void impulse_mutex_unlock(impulse_mutex mutex)
{
  THREAD_ERROR(pth_mutex_release(mutex));
}

void impulse_mutex_destroy(impulse_mutex mutex)
{
  impulse_free(mutex);
}

static pth_mutex_t global_lock = PTH_MUTEX_INIT;

void impulse_mutex_f_global_lock(void)
{
  THREAD_ERROR(pth_mutex_acquire(&global_lock, FALSE, NULL));
}

void impulse_mutex_f_global_unlock(void)
{
  THREAD_ERROR(pth_mutex_release(&global_lock));
}

/*                            ======--------======                           */

/* Condition variable functions */

impulse_cond impulse_cond_create(void)
{
  pth_cond_t *cond;

  cond = impulse_alloc(sizeof(pth_cond_t));

  THREAD_ERROR(pth_cond_init(cond));
  
  return cond;
}

void impulse_cond_signal(impulse_cond  cond)
{
  THREAD_ERROR(pth_cond_notify(cond, FALSE));
}

void impulse_cond_broadcast(impulse_cond  cond)
{
  THREAD_ERROR(pth_cond_notify(cond, TRUE));
}

void impulse_cond_wait(impulse_cond  cond,
		       impulse_mutex mutex)
{
  THREAD_ERROR(pth_cond_await(cond, mutex, NULL));
}

void impulse_cond_destroy(impulse_cond cond)
{
  impulse_free(cond);
}

/*                            ======--------======                           */

/* Read/write mutexes */

impulse_rwlock impulse_rwlock_create(void)
{
  impulse_rwlock lock;

  lock = impulse_alloc(sizeof(pth_rwlock_t));

  THREAD_ERROR(pth_rwlock_init(lock));

  return lock;
}


void impulse_rwlock_read(impulse_rwlock lock)
{
  THREAD_ERROR(pth_rwlock_acquire(lock, PTH_RWLOCK_RD, FALSE, NULL));
}


impulse_bool impulse_rwlock_tryread(impulse_rwlock lock)
{
  impulse_int ret;
  ret = pth_rwlock_acquire(lock, PTH_RWLOCK_RD, TRUE, NULL);
  if (ret == EBUSY)
    return 0;
  else if (ret < 0)
    THREAD_ERROR(ret);
  return 1;
}


void impulse_rwlock_write(impulse_rwlock lock)
{
  THREAD_ERROR(pth_rwlock_acquire(lock, PTH_RWLOCK_RW, FALSE, NULL));
}


impulse_bool impulse_rwlock_trywrite(impulse_rwlock lock)
{
  impulse_int ret;
  ret = pth_rwlock_acquire(lock, PTH_RWLOCK_RW, TRUE, NULL);
  if (ret == EBUSY)
    return 0;
  else if (ret < 0)
    THREAD_ERROR(ret);
  return 1;
}


void impulse_rwlock_unlock(impulse_rwlock lock)
{
  THREAD_ERROR(pth_rwlock_release(lock));
}


void impulse_rwlock_destroy(impulse_rwlock lock)
{
  impulse_free(lock);
}


/* Semaphores */


impulse_sema impulse_sema_create(impulse_int value)
{
  impulse_sema sem = impulse_alloc(sizeof(impulse_sema_impl));
  
  THREAD_ERROR(pth_mutex_init(&sem->mutex));
  THREAD_ERROR(pth_cond_init(&sem->cond));
  sem->value = value;
  
  return sem;
}


void impulse_sema_wait(impulse_sema sem)
{
  THREAD_ERROR(pth_mutex_acquire(&sem->mutex, FALSE, NULL));
  
  if (sem->value > 0)
  {
    sem->value--;
  }
  else
  {
    THREAD_ERROR(pth_cond_await(&sem->cond, &sem->mutex, NULL));
  }
  
  THREAD_ERROR(pth_mutex_release(&sem->mutex));
}


impulse_bool impulse_sema_trywait(impulse_sema sem)
{
  THREAD_ERROR(pth_mutex_acquire(&sem->mutex, FALSE, NULL));
  
  if (sem->value > 0)
  {
    sem->value--;
    THREAD_ERROR(pth_mutex_release(&sem->mutex));
    return 1;
  }
  else
  {
    THREAD_ERROR(pth_mutex_release(&sem->mutex));
    return 0;
  }
}


void impulse_sema_post(impulse_sema sem)
{
  THREAD_ERROR(pth_mutex_acquire(&sem->mutex, FALSE, NULL));
  
  if (sem->value == 0)
    pth_cond_notify(&sem->cond, 0);
  else
    sem->value++;
  
  THREAD_ERROR(pth_mutex_release(&sem->mutex));
}


void impulse_sema_destroy(impulse_sema sem)
{
  impulse_free(sem);
}

/* End of file */
