/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * unidata.h
 * ---------
 *
 * Data structures used to represent the Unicode Character Database
 * (www.unicode.org), needed for correct manipulation of Unicode
 * characters. Also structures for representing conversion tables
 * between Unicode and arbitary 
 */

#ifndef __impulse_util_unidata_h
#define __impulse_util_unidata_h

#include <impulse/util/unichar.h>

/*                            ======--------======                           */

/* Unicode version prevalent when this file was written */

#define UNICODE_MAJOR 2
#define UNICODE_MINOR 1

/*                            ======--------======                           */

/* Category flags */

/* Normative */

#define UNICODE_NONSPACING_MARK       1
#define UNICODE_COMBINING_MARK        2
#define UNICODE_NUMBER_DECIMAL        4
#define UNICODE_NUMBER_OTHER          8
#define UNICODE_SEPARATOR_SPACE      16
#define UNICODE_SEPARATOR_LINE       32
#define UNICODE_SEPARATOR_PARAGRAPH  64
#define UNICODE_OTHER_CONTROL       128
#define UNICODE_OTHER_PRIVATE       256
#define UNICODE_OTHER_NOT_ASSIGNED  512

/* Informative */

#define UNICODE_LETTER_UPPERCASE      1
#define UNICODE_LETTER_LOWERCASE      2
#define UNICODE_LETTER_TITLECASE      4
#define UNICODE_LETTER_MODIFIER       8
#define UNICODE_LETTER_OTHER         16
#define UNICODE_PUNCTUATION_DASH     32
#define UNICODE_PUNCTUATION_OPEN     64
#define UNICODE_PUNCTUATION_CLOSE   128
#define UNICODE_PUNCTUATION_OTHER   256
#define UNICODE_SYMBOL_MATH         512
#define UNICODE_SYMBOL_CURRENCY    1024
#define UNICODE_SYMBOL_OTHER       2048

/* Bidirectional information */

typedef enum {
  UNICODE_LEFT_RIGHT = 0,
  UNICODE_RIGHT_LEFT,

  /* Weak types */
  UNICODE_EURO_NUMBER,
  UNICODE_EURO_NUMBER_SEPARATOR,
  UNICODE_EURO_NUMBER_TERMINATOR,
  UNICODE_ARABIC_NUMBER,
  UNICODE_COMMON_NUMBER_SEPARATOR,

  /* Separators */
  UNICODE_BLOCK_SEPARATOR,
  UNICODE_SEGMENT_SEPARATOR,

  /* Neutrals */
  UNICODE_NEUTRAL_WHITESPACE,
  UNICODE_NEUTRAL_OTHER
} impulse_unicode_direction;

/* Canonical combining classes */

#define UNICODE_COMBINING_SPACING                0
#define UNICODE_COMBINING_OVERLAY                1
#define UNICODE_COMBINING_TIBETAN_SUBJOINED      6
#define UNICODE_COMBINING_NUKTAS                 7
#define UNICODE_COMBINING_HIRAGANA               8
#define UNICODE_COMBINING_KATAKANA               8
#define UNICODE_COMBINING_VIRAMAS                9
#define UNICODE_COMBINING_START_FIXED           10
#define UNICODE_COMBINING_END_FIXED            199
#define UNICODE_COMBINING_BELOW_LEFT_ATTACHED  200
#define UNICODE_COMBINING_BELOW_ATTACHED       202
#define UNICODE_COMBINING_BELOW_RIGHT_ATTACHED 204
#define UNICODE_COMBINING_LEFT_ATTACHED        208
#define UNICODE_COMBINING_RIGHT_ATTACHED       210
#define UNICODE_COMBINING_ABOVE_LEFT_ATTACHED  212
#define UNICODE_COMBINING_ABOVE_ATTACHED       214
#define UNICODE_COMBINING_ABOVE_RIGHT_ATTACHED 216
#define UNICODE_COMBINING_BELOW_LEFT           218
#define UNICODE_COMBINING_BELOW                220
#define UNICODE_COMBINING_BELOW_RIGHT          222
#define UNICODE_COMBINING_LEFT                 224
#define UNICODE_COMBINING_RIGHT                226
#define UNICODE_COMBINING_ABOVE_LEFT           228
#define UNICODE_COMBINING_ABOVE                230
#define UNICODE_COMBINING_ABOVE_RIGHT          232
#define UNICODE_COMBINING_DOUBLE_ABOVE         234

/* Formatting tags */

#define UNICODE_FORMAT_FONT                      1
#define UNICODE_FORMAT_NOBREAK                   2
#define UNICODE_FORMAT_INITIAL                   4
#define UNICODE_FORMAT_MEDIAL                    8
#define UNICODE_FORMAT_FINAL                    16
#define UNICODE_FORMAT_ISOLATED                 32
#define UNICODE_FORMAT_CIRCLE                   64
#define UNICODE_FORMAT_SUPER                   128
#define UNICODE_FORMAT_SUB                     256
#define UNICODE_FORMAT_VERTICAL                512
#define UNICODE_FORMAT_WIDE                   1024
#define UNICODE_FORMAT_NARROW                 2048
#define UNICODE_FORMAT_SMALL                  4096
#define UNICODE_FORMAT_SQUARE                 8192
#define UNICODE_FORMAT_COMPAT                16384

/* Unidata structures */

/*
 * FIXME: as there is apparently no provision for surrogate pairs in
 * the Unicode Character Database file format, no provision is made
 * here.
 */

/*
 * Fields marked as optional may not be stored for reasons of space or 
 * omission from the Unicode Character Database. Omitted fields will
 * be NULL (the pointer value or the Unicode value [U+0000]).
 */

/*
 * I've done my best to minimize stored database size, but I assume
 * gcc semantics when it comes to packing structures (gcc could
 * definately be improved in this area - it should automatically
 * rearrange structures to put similar typed elements together). Even
 * so, you can expect a 450k+ object file with the full Unicode
 * database. Not good for static linking ;-)
 */

typedef struct {
  /* "Name<normative>;Oldname<informative>;Comment<informative>"      */
  char                     *info;            /* Optional              */
  impulse_short                     category_n;      /* Flags, normative      */
  impulse_short                     category_i;      /* Flags, informative    */

  impulse_unichar           uppercase;       /* Informative, optional */
  impulse_unichar           lowercase;       /* Informative, optional */
  impulse_unichar           titlecase;       /* Informative, optional */

  impulse_ubyte                      combining_class; /* Normative             */
  impulse_ubyte                      direction;       /* Normative             */
  impulse_ubyte                      decimal_value;   /* Normative             */
  impulse_ubyte                      is_mirrored;     /* Normative, boolean    */

  struct {
    impulse_ubyte     length;
    impulse_short    formatting;
    impulse_unichar *decomposition;
  }                         decomposition;   /* Normative             */
  
  /* impulse_int                     digit_value;     Informative           */
  float                     numeric_value;   /* Normative, approx     */
} impulse_unidata_item;

/*                            ======--------======                           */

/* Assorted interesting character ranges */

#define UNICODE_SURROGATEHIGH_START (0xD800)
#define UNICODE_SURROGATEHIGH_END   (0xDB7F)
#define UNICODE_SURROGATELOW_START  (0xDC00)
#define UNICODE_SURROGATELOW_END    (0xDFFF)

/*                            ======--------======                           */

/*
 * Encoded form of Unicode Character Data (256 arrays of 256 entries
 * each)
 */

extern impulse_unidata_item *impulse_unidata[];

/*                            ======--------======                           */

/* Composition table structures */

#define IMPULSE_UNIDATA_COMPO_BINARY 1
#define IMPULSE_UNIDATA_COMPO_MULTI  2
#define IMPULSE_UNIDATA_COMPO_UNUSED 4

typedef struct _impulse_unidata_composition
{
  impulse_ubyte flags;     /* Flags */
  impulse_unichar composition;

  int number_multis;
  impulse_unichar   *multi_compo;
  struct _multi_data
  {
    impulse_ubyte length;
    impulse_unichar *string;
  } *multi;
} impulse_unidata_composition;

/* Hey! That's a lot of asterisks! (15, in fact) */
extern impulse_unidata_composition ***************impulse_unidata_comp_table[4];


#endif
